/******************************************************************************
  Copyright (c) 2000-2007 by Turku PET Centre

  File:        dft.c
  Description: Contains functions for reading and writing DFT files and
               processing TAC data.
  Known bugs:  Line length cannot be larger than max int value, but this is not
               checked. This may limit the curve number in some systems.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi

  Modification history:
  2000-07-07 Vesa Oikonen
    First created.
  2000-07-26 VO
    _type is set to 0 (not 2) in case of "plain" data without titles.
    For plain datafiles only frame mid times are saved.
  2000-08-07 VO
    Checks for overlapping and very small gaps in frametimesDFT().
  2000-09-09 VO
    Added place for weights in the data structure.
    Weights are saved as a VOI in DFT files.
    Included function copyvoiDFT().
  2000-09-18 VO
    Weights are not saved with "plain" data.
  2000-09-21 VO
    Included function addnullframeDFT().
  2000-11-30 VO
    Fixed a bug in addnullframeDFT().
  2000-12-13 VO
    Minor changes in source code, not affecting the compiled code.
  2001-01-12 VO
    Included function sortDFT().
    addnullframeDFT(): x2[0]=x1[1]
  2001-01-24 VO
    Extracting study number works as intended with MSDOS pathnames.
  2001-01-30 VO
    selectDFT() does not change its argument 'name'.
  2001-01-31 VO
    Additional data is initialized in addmemDFT().
  2001-03-25 VO
    Included functions copymainhdrDFT() and copyvoihdrDFT().
  2001-03-31 VO
    Fixed a bug in copymainhdrDFT().
  2001-04-10 VO
    Minor change in selectDFT().
  2001-04-26 VO
    overflowDFT() returns now different error numbers.
  2001-10-30 VO
    Recognizes specifically DFT fit format (but does not read).
  2002-02-21 VO
    addDFT() allows different units if _type is 0.
  2002-07-30 VO
    memset() added to initDFT().
  2002-10-10 VO
    sortDFT() rewritten.
  2002-12-01 VO
    Change in frametimesDFT(); now works with just one frame.
  2003-02-12 VO
    Change in readDFT(); ignores data line after comment sign '#'.
  2003-06-19 VO
    Included function na_fillDFT().
  2003-08-13 VO
    Included "timetypes" um and mm for measuring distances.
    Strings "Distance" and "Distances" and units "um", "mm" and s are
    identified in addition to the previous "Time" and "Times" and
    "sec" and "min".
  2003-08-14 VO
    Unknown timetype is read/written as string "()".
  2003-12-05 KS & VO
    Included functions kBqMinDFT() and kBqMaxDFT().
  2004-01-21 VO
    Returns error as it should when trying to read empty file.
  2004-01-26 VO
    Included function dftdup().
    dftadd() applies copyvoihdrDFT().
    Function infos are changed.
  2004-01-27 VO
    Included function dftMovevoi().
  2004-05-23 VO
    Changed some comments into Doxygen style.
    Included function dftSortByFrame().
  2004-05-25 VO
    Included function dftWriteHTML().
  2004-08-17 VO
    Included function dftDelete().
  2004-08-23 VO
    Function names with *DFT end changed to start with dft*.
    New library function studynr_from_fname() is applied.
  2004-08-26 VO
    Unit check in dftAdd() is changed to be case-insensitive.
  2004-09-05 VO
    Reads and copies and prints voi.name but does not yet write it.
    Included function dftSelectRegions().
  2004-09-20 VO
    Doxygen style comments corrected.
  2004-12-04 VO
    dftRead() tries to read also old *.roi.kbq / *.roi.nci files.
  2004-12-09 VO
    Included function dftSortByPlane().
  2005-06-17 VO
    Added one digit in printing floats in exponential format because
    of MinGW compiler.
  2005-10-30 VO
    New fields in DFT are initiated and copied when necessary.
    Included function dft_fill_hdr_from_IFT().
    dftRead():
      Tries to read header info in comments.
    dftWrite():
      Writes newline always before '#' in comments.
      Writes HTML format if filename extension is *.HTM(L).
    dftWriteHTML():
      Big changes in HTML codes, but table looks as before.
  2005-11-30 VO
    dftRead():
      Tries to read IDWC and IF file, based on filename extension.
  2005-12-05 Kaisa Sederholm
    dftWrite():
      Added warning if the function is writing the file with a name
      that allready exists.
  2006-02-01 VO
    "allready" -> "already".
  2006-05-15 VO
    Corrected bugs in dftFrametimes() which changed time unit
    when sample number was 1.
  2006-06-03 VO
    Some strcmp() -> strcasecmp().
    dftRead(): atof() replaced by decimal point insensitive atof_dpi().
  2006-07-12 VO
    Added functions dft_nr_of_NA() and dftMinMax().
  2006-12-01 VO
    dftWrite(), when file exists:
      If backup file exists, it is removed before attempting rename.
      Removed warning about existing file, because it is too common and
      leads to ignoring more important warnings as well.
  2007-01-12 VO
    Included function dftDeleteFrameOverlap(DFT *dft).
    dftWriteHTML() includes in HTML code Excel specific comments specifying
      that this file has points as decimal separator.
    HTML table format changed to valid XHTML 1.1.
    Added functions dftWriteXHTML11_doctype() and dftWriteXHTML11_head().
  2007-01-26 VO
    Bug correction in dftWriteHTML(): <table> was not written if column titles
    were not written.
  2007-02-10 VO
    Functions dftMin2sec() and dftSec2min were moved to a new file dftunit.c.
    I/O functions were moved to a new file dftio.c.
    dftDeleteFrameOverlap(): time overlap limit changed from 1 sec to 1.8 sec.
    Uses new unit enums.
  2007-04-13 VO
    Added function dftSetComments(DFT *dft).
  2007-08-09 VO
    Function dftDeleteFrameOverlap() mostly rewritten; old version saved as
    dftDeleteFrameOverlap_old().
  2007-08-13 VO
    Uses new unit enums and functions.


******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
/*****************************************************************************/
#include "decpoint.h"
/*****************************************************************************/
#include "include/dft.h"
#include "include/dftunit.h"
#include "include/dftio.h"
/* local function definitions */
int dftQSortName(const void *voi1, const void *voi2);
int dftQSortPlane(const void *voi1, const void *voi2);
/*****************************************************************************/

/*****************************************************************************/
/** Free memory allocated for DFT. All data is cleared. */
void dftEmpty(DFT *data)
{
  if(data->_voidataNr>0) free((char*)(data->voi));
  if(data->_dataSize>0) free((char*)(data->_data));
  data->_dataSize=data->_voidataNr=0;
  data->frameNr=data->voiNr=0;
  data->studynr[0]=data->comments[0]=data->unit[0]=(char)0;
  data->radiopharmaceutical[0]=data->isotope[0]=data->decayCorrected=(char)0;
  data->scanStartTime[0]=data->injectionTime[0]=(char)0;
  data->timeunit=data->timetype=0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Initiate DFT structure. This should be called once before use. */
void dftInit(DFT *data)
{
  memset(data, 0, sizeof(DFT));
  data->_voidataNr=data->_dataSize=data->_dataSize=0;
  data->frameNr=data->voiNr=0;
  data->studynr[0]=data->comments[0]=data->unit[0]=(char)0;
  data->radiopharmaceutical[0]=data->isotope[0]=data->decayCorrected=(char)0;
  data->scanStartTime[0]=data->injectionTime[0]=(char)0;
  data->timeunit=data->timetype=data->isweight=0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Allocate memory for DFT data and sets data pointers. Old data is destroyed.
 */
int dftSetmem(DFT *data, int frameNr, int voiNr)
{
  int i, n;
  double *d;


  /* Clear previous data */
  dftEmpty(data);

  /* Allocate memory for curves */
  data->voi=(Voi*)calloc(voiNr, sizeof(Voi));
  if(data->voi==NULL) return 1;
  data->_voidataNr=voiNr;

  /* Allocate memory for frames */
  /* For 3 axes, weights, and for curves (3 for each VOI) */
  /* And one extra 'frame' for overflow testing */
  n=(frameNr+1)*(3+1+3*voiNr);
  data->_data=(double*)calloc(n, sizeof(double));
  if(data->_data==NULL) return 1;
  data->_dataSize=n;

  /* Set pointers for curve data */
  d=data->_data; data->x = d;
  d += frameNr + 1; data->x1 = d;
  d += frameNr + 1; data->x2 = d;
  d += frameNr + 1; data->w = d;
  for(i=0; i<voiNr; i++) {
    d += frameNr + 1; data->voi[i].y = d;
    d += frameNr + 1; data->voi[i].y2 = d;
    d += frameNr + 1; data->voi[i].y3 = d;
  }

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Allocate memory for specified nr of additional data VOIs.
    Old data is left unchanged.
 */
int dftAddmem(DFT *data, int voiNr)
{
  int i, j, n;
  double *d;


  /* Allocate memory for curves */
  data->voi=(Voi*)realloc(data->voi, (voiNr+data->_voidataNr)*sizeof(Voi));
  if(data->voi==NULL) return 1;
  data->_voidataNr+=voiNr;

  /* Allocate memory for frames */
  /* For 3 axes, weights, and for curves (3 for each VOI) */
  /* And one extra 'frame' for overflow testing */
  n=(data->frameNr+1)*(3*voiNr);
  data->_data=(double*)realloc(data->_data, (n+data->_dataSize)*sizeof(double));
  if(data->_data==NULL) return 1;
  data->_dataSize+=n;

  /* Set pointers for curve data */
  d=data->_data; data->x = d; data->x[data->frameNr]=0.0;
  d += data->frameNr + 1; data->x1 = d; data->x1[data->frameNr]=0.0;
  d += data->frameNr + 1; data->x2 = d; data->x2[data->frameNr]=0.0;
  d += data->frameNr + 1; data->w = d; data->w[data->frameNr]=0.0;
  for(i=0; i<data->_voidataNr; i++) {
    d += data->frameNr + 1; data->voi[i].y = d;
    d += data->frameNr + 1; data->voi[i].y2 = d;
    d += data->frameNr + 1; data->voi[i].y3 = d;
  }

  /* Initiate values */
  for(i=data->voiNr; i<data->_voidataNr; i++) {
    strcpy(data->voi[i].name, "");
    strcpy(data->voi[i].voiname, "");
    strcpy(data->voi[i].hemisphere, "");
    strcpy(data->voi[i].place, "");
    data->voi[i].size=1.0;
    data->voi[i].sw=data->voi[i].sw2=data->voi[i].sw3=0;
    for(j=0; j<data->frameNr+1; j++)
      data->voi[i].y[j]=data->voi[i].y2[j]=data->voi[i].y3[j]=0.0;
  }

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Add the specified voi [0,voiNr-1] from data2 to data1.
    Allocates memory for additional data VOI, if necessary.
\return Returns 0 if ok.
 */
int dftAdd(DFT *data1, DFT *data2, int voi)
{
  int i, n;

  /* Check that voi exists */
  if(data2->voiNr<=voi || voi<0) {
    strcpy(dfterrmsg, "there is no region to combine"); return 8;}

  /* Check that frame number etc is the same */
  if(data1->frameNr!=data2->frameNr ||
     (data1->_type>0 && data2->_type>0 &&
      (data1->timeunit!=data2->timeunit || strcasecmp(data1->unit, data2->unit))
   )) {
    strcpy(dfterrmsg, "data does not match"); return 8;}

  /* Allocate more memory if necessary */
  if(data1->_voidataNr==data1->voiNr)
    if(dftAddmem(data1, 1)) {
      strcpy(dfterrmsg, "cannot allocate memory"); return 8;}

  /* Copy data */
  n=data1->voiNr;
  (void)dftCopyvoihdr(data2, voi, data1, n);
  for(i=0; i<data1->frameNr; i++) {
    data1->voi[n].y[i]=data2->voi[voi].y[i];
    data1->voi[n].y2[i]=data2->voi[voi].y2[i];
    data1->voi[n].y3[i]=data2->voi[voi].y3[i];
  }
  data1->voiNr+=1;

  /* If data2 contains weights and data1 does not, then copy those */
  if(data2->isweight && !data1->isweight)
    for(i=0; i<data1->frameNr; i++) data1->w[i]=data2->w[i];

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Select VOIs (sets sw=1), whose names are matching specified string.
    If no string is specified, then all VOIs are selected.
\return Returns the number of matches, or <0, if an error occurred.
 */
int dftSelect(DFT *data, char *name)
{
  int i, j, n;
  char *p, n1[128], n2[128], n3[128], tmp[128], sname[1024];

  /* Select all, if no string was specified */
  if(name==NULL || strlen(name)==0) {
    for(i=0; i<data->voiNr; i++) data->voi[i].sw=1; return data->voiNr;}
  /* Make a copy of 'name' and use it */
  strcpy(sname, name);
  /* Check if string contains several substrings (hemisphere and place) */
  n1[0]=n2[0]=n3[0]=(char)0;
  p=strtok(sname, " ,;\n\t|"); if(p!=NULL) strcpy(n1, p); else return -1;
  p=strtok(NULL, " ,;\n\t|"); if(p!=NULL) {
    strcpy(n2, p); p=strtok(NULL, " ,;\n\t|"); if(p!=NULL) strcpy(n3, p);}
  /* Convert strings to lowercase */
  for(i=0; i<strlen(n1); i++) n1[i]=tolower(n1[i]);
  for(i=0; i<strlen(n2); i++) n2[i]=tolower(n2[i]);
  for(i=0; i<strlen(n3); i++) n3[i]=tolower(n3[i]);
  /* Search through the data */
  for(i=0, n=0; i<data->voiNr; i++) {
    data->voi[i].sw=0;
    snprintf(tmp, 128, "%s%s%s", data->voi[i].voiname, data->voi[i].hemisphere,
                                 data->voi[i].place);
    for(j=0; j<strlen(tmp); j++) tmp[j]=tolower(tmp[j]);
    if(strstr(tmp, n1)==NULL) continue;
    if(n2[0] && strstr(tmp, n2)==NULL) continue;
    if(n3[0] && strstr(tmp, n3)==NULL) continue;
    data->voi[i].sw=1; n++;
  }
  return n;
}
/*****************************************************************************/

/*****************************************************************************/
/** Select the VOIs that have matching region name or number.
    Sets sw=1 or sw=0. This will replace dftSelect().
\return Returns the number of selected VOIs, or <0 in case of an error.
 */
int dftSelectRegions(
  /** Pointer to DFT data where VOIs are selected */
  DFT *dft,
  /** Name or VOI number which is searched */
  char *region_name,
  /** 1=Non-matching VOIs are deselected, 0=Old selections are preserved */
  int reset
) {
  int ri, match_nr=0;

  /* Check the input */
  if(dft==NULL || dft->voiNr<1 || strlen(region_name)<1) return(-1);
  /* Reset all selections if required */
  if(reset!=0) for(ri=0; ri<dft->voiNr; ri++) dft->voi[ri].sw=0;
  /* Check each VOI */
  for(ri=0; ri<dft->voiNr; ri++) {
    if(rnameMatch(dft->voi[ri].name, ri+1, region_name)!=0) {
      dft->voi[ri].sw=1; match_nr++;
    }
  }
  return(match_nr);
}
/*****************************************************************************/

/*****************************************************************************/
/** Calculate frame mid or start and end times. Timetype is not changed. */
void dftFrametimes(DFT *data)
{
  int i, j, sw=0;
  double f, fs;

  /* Do this only if timeunit is known to be min or sec */
  if(data->timeunit!=TUNIT_SEC && data->timeunit!=TUNIT_MIN) return;

  /* If times are in min, change to sec to allow rounding to ints */
  if(data->timeunit==TUNIT_MIN) {sw=1; dftMin2sec(data);}

  /* Decide what to do, and get it done */
  if(data->timetype==0) { /* frame start and end times from mid times */
    /* Easy, if only one frame */
    if(data->frameNr==1) {
      if(data->x[0]<=0.0) {data->x1[0]=data->x[0]; data->x2[0]=0.0;}
      else {data->x1[0]=0.0; data->x2[0]=2.0*data->x[0];}
      if(sw) dftSec2min(data);
      return;
    }
    /* Fill start and end times with -999 */
    for(i=0; i<data->frameNr; i++) data->x1[i]=data->x2[i]=-999.;
    /* Search for sequences of nearly same frame lengths */
    for(i=1; i<data->frameNr-1; i++) {
      f=data->x[i]-data->x[i-1]; fs=data->x[i+1]-data->x[i];
      if(fabs(f-fs)>=2.0) continue; else f=(f+fs)/2.0;
      data->x1[i-1]=data->x[i-1]-f/2.0; data->x2[i-1]=data->x[i-1]+f/2.0;
      data->x1[i]=data->x[i]-f/2.0; data->x2[i]=data->x[i]+f/2.0;
      data->x1[i+1]=data->x[i+1]-f/2.0; data->x2[i+1]=data->x[i+1]+f/2.0;
      /* Check for negatives */
      for(j=i-1; j<i+2; j++) {
        if(data->x1[j]<0.0) data->x1[j]=0.0;
        if(data->x2[j]<0.0) data->x2[j]=0.0;
      }
    }
    /* If out-of-sequence frames were left out, fill those to the nearest one */
    i=0;  /* first frame */
    if(data->x1[i]<0) {
      if(data->x1[i+1]>0) data->x2[i]=data->x1[i+1];
      else data->x2[i]=(data->x[i+1]+data->x[i])/2.0;
      data->x1[i]=2.0*data->x[i]-data->x2[i];
    }
    i=data->frameNr-1;  /* last frame */
    if(data->x1[i]<0) {
      if(data->x2[i-1]>0) data->x1[i]=data->x2[i-1];
      else data->x1[i]=(data->x[i-1]+data->x[i])/2.0;
      data->x2[i]=2.0*data->x[i]-data->x1[i];
    }
    /* other frames */
    for(i=1; i<data->frameNr-1; i++) if(data->x1[i]<0.0) {
      /* which frame is nearest? */
      if(data->x[i]-data->x[i-1] < data->x[i+1]-data->x[i]) { /* last one */
        if(data->x2[i-1]>0) data->x1[i]=data->x2[i-1];
        else data->x1[i]=(data->x[i-1]+data->x[i])/2.0;
        data->x2[i]=2.*data->x[i]-data->x1[i];
      } else { /* next one */
        if(data->x1[i+1]>0) data->x2[i]=data->x1[i+1];
        else data->x2[i]=(data->x[i+1]+data->x[i])/2.0;
        data->x1[i]=2.0*data->x[i]-data->x2[i];
      }
    }
    /* Check for negatives */
    for(i=0; i<data->frameNr; i++) {
      if(data->x1[i]<0.0) data->x1[i]=0.0;
      if(data->x2[i]<0.0) data->x2[i]=data->x1[i];
    }
    /* Check for overlapping and very small gaps */
    for(i=1; i<data->frameNr; i++) {
      f=data->x1[i]-data->x2[i-1];
      if(f<1.0) data->x1[i]=data->x2[i-1]=(data->x1[i]+data->x2[i-1])/2.0;
    }
  } else if(data->timetype==3) { /* mid times from frame start and end times */
    for(i=0; i<data->frameNr; i++) data->x[i]=0.5*(data->x1[i]+data->x2[i]);
  } else if(data->timetype==1) { /* frame start times -> end and mid times */
    for(i=0; i<data->frameNr-1; i++) data->x2[i]=data->x1[i+1];
    data->x2[data->frameNr-1]=data->x1[data->frameNr-1]+
      (data->x2[data->frameNr-2]-data->x1[data->frameNr-2]);
    for(i=0; i<data->frameNr; i++) data->x[i]=0.5*(data->x1[i]+data->x2[i]);
  } else if(data->timetype==2) { /* frame end times -> start and mid times */
    data->x1[0]=0.0;
    for(i=1; i<data->frameNr; i++) data->x1[i]=data->x2[i-1];
    for(i=0; i<data->frameNr; i++) data->x[i]=0.5*(data->x1[i]+data->x2[i]);
  }

  /* If times were in min, change them back */
  if(sw) dftSec2min(data);

  return;
}
/*****************************************************************************/

/*****************************************************************************/
/** Check for overflows in data structure. Returns 0, if ok. */
int dftOverflow(DFT *data)
{
  int i;

  if(data->frameNr<1 || data->voiNr<1) return 0;
  if(data->x[data->frameNr]!=0.0) return 1;
  if(data->x1[data->frameNr]!=0.0) return 2;
  if(data->x2[data->frameNr]!=0.0) return 3;
  for(i=0; i<data->voiNr; i++) {
    if(data->voi[i].y[data->frameNr]!=0.0) return 4;
    if(data->voi[i].y2[data->frameNr]!=0.0) return 5;
    if(data->voi[i].y3[data->frameNr]!=0.0) return 6;
  }
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Copy VOI data inside DFT data structure from one place to another. */
int dftCopyvoi(DFT *data, int from, int to)
{
  int i;

  /* Check that required data exists */
  if(to>=data->_voidataNr || from>=data->_voidataNr) return 1;
  if(from==to) return 0;

  /* Copy VOI info */
  strcpy(data->voi[to].name, data->voi[from].name);
  strcpy(data->voi[to].voiname, data->voi[from].voiname);
  strcpy(data->voi[to].hemisphere, data->voi[from].hemisphere);
  strcpy(data->voi[to].place, data->voi[from].place);
  data->voi[to].size=data->voi[from].size;
  data->voi[to].sw=data->voi[from].sw;
  data->voi[to].sw2=data->voi[from].sw2;
  data->voi[to].sw3=data->voi[from].sw3;
  /* Copy VOI curves */
  for(i=0; i<data->frameNr; i++) {
    data->voi[to].y[i]=data->voi[from].y[i];
    data->voi[to].y2[i]=data->voi[from].y2[i];
    data->voi[to].y3[i]=data->voi[from].y3[i];
  }

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Move VOI in DFT structure from one position to another. */
int dftMovevoi(DFT *dft, int from, int to)
{
  int ri;
  size_t voisize;
  Voi voi;

  if(from<0 || to<0) return(1);
  if(from+1>dft->_voidataNr || to+1>dft->_voidataNr) return(2);
  if(from==to) return(0);
  voisize=sizeof(Voi);
  memcpy(&voi, dft->voi+from, voisize);
  if(from>to) for(ri=from; ri>to; ri--)
    memcpy(dft->voi+ri, dft->voi+(ri-1), voisize);
  else for(ri=from; ri<to; ri++)
    memcpy(dft->voi+ri, dft->voi+(ri+1), voisize);
  memcpy(dft->voi+ri, &voi, voisize);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Delete specified TAC (0..voiNr-1) from the DFT structure.
\return Returns 0 if ok.
 */
int dftDelete(DFT *dft, int voi)
{
  int i, ret;

  /* Check that region exists */
  if(voi>dft->voiNr-1 || voi<0) return(1);
  /* If it is the last one, then just decrease the voiNr */
  if(voi==dft->voiNr) {dft->voiNr--; return(0);}
  /* Otherwise we have to move the following regions in its place */
  for(i=voi+1; i<dft->voiNr; i++) {
    ret=dftMovevoi(dft, i, i-1); if(ret) return(10+ret);
  }
  dft->voiNr--;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Copy main header info from dft1 to dft2. */
int dftCopymainhdr(DFT *dft1, DFT *dft2)
{
  strcpy(dft2->studynr, dft1->studynr);
  strcpy(dft2->unit, dft1->unit);
  dft2->timeunit=dft1->timeunit; dft2->timetype=dft1->timetype;
  strcpy(dft2->comments, dft1->comments);
  strcpy(dft2->radiopharmaceutical, dft1->radiopharmaceutical);
  strcpy(dft2->isotope, dft1->isotope);
  strcpy(dft2->scanStartTime, dft1->scanStartTime);
  strcpy(dft2->injectionTime, dft1->injectionTime);
  dft2->decayCorrected=dft1->decayCorrected;
  dft2->_type=dft1->_type;
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Copy voi header info from dft1.voi[from] to dft2.voi[to]. */
int dftCopyvoihdr(DFT *dft1, int from, DFT *dft2, int to)
{
  /* Check that required data exists */
  if(to>=dft2->_voidataNr || from>=dft1->_voidataNr) return 1;

  /* Copy VOI info */
  strcpy(dft2->voi[to].name, dft1->voi[from].name);
  strcpy(dft2->voi[to].voiname, dft1->voi[from].voiname);
  strcpy(dft2->voi[to].hemisphere, dft1->voi[from].hemisphere);
  strcpy(dft2->voi[to].place, dft1->voi[from].place);
  dft2->voi[to].size=dft1->voi[from].size;
  dft2->voi[to].sw=dft1->voi[from].sw;
  dft2->voi[to].sw2=dft1->voi[from].sw2;
  dft2->voi[to].sw3=dft1->voi[from].sw3;

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Makes a duplicate of DFT structure pointed to by dft1 into dft2.
    Any existing content of dft2 will be deleted. Dft2 must be initiated.
\return Returns 0 if ok.
 */
int dftdup(DFT *dft1, DFT *dft2)
{
  int ri, fi, ret;

  /* Empty the new data */
  dftEmpty(dft2);
  /* Is that it? Is there any contents in dft1? */
  if(dft1->voiNr==0 && dft1->frameNr==0) {
    ret=dftCopymainhdr(dft1, dft2); return(ret);
  }
  /* Allocate memory for dft2 */
  ret=dftSetmem(dft2, dft1->frameNr, dft1->voiNr); if(ret) return(ret);
  dft2->voiNr=dft1->voiNr; dft2->frameNr=dft1->frameNr;
  /* Copy the contents */
  ret=dftCopymainhdr(dft1, dft2); if(ret) return(ret);
  for(ri=0; ri<dft1->voiNr; ri++) {
    ret=dftCopyvoihdr(dft1, ri, dft2, ri); if(ret) return(ret);
    for(fi=0; fi<dft1->frameNr; fi++) {
      dft2->voi[ri].y[fi]=dft1->voi[ri].y[fi];
      dft2->voi[ri].y2[fi]=dft1->voi[ri].y2[fi];
      dft2->voi[ri].y3[fi]=dft1->voi[ri].y3[fi];
    }
  }
  for(fi=0; fi<dft1->frameNr; fi++) {
    dft2->x[fi]=dft1->x[fi];
    dft2->x1[fi]=dft1->x1[fi]; dft2->x2[fi]=dft1->x2[fi];
    dft2->w[fi]=dft1->w[fi];
  }
  dft2->isweight=dft1->isweight;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Include a frame with time 0, unless one already exists. */
int dftAddnullframe(DFT *data)
{
  int i, j, n;
  DFT temp;


  /* Check whether nullframe exists */
  if(data->x[0]==0.0) return 0;

  /* Allocate memory for temp data */
  dftInit(&temp);
  if(dftSetmem(&temp, data->frameNr, data->voiNr)) return 1;
  temp.frameNr=data->frameNr; temp.voiNr=data->voiNr;

  /* Copy data to temp */
  strcpy(temp.unit, data->unit);
  temp.timeunit=data->timeunit; temp.timetype=data->timetype;
  temp.isweight=data->isweight; temp._type=data->_type;
  strcpy(temp.comments, data->comments);
  for(j=0; j<data->frameNr; j++) {
    temp.x[j]=data->x[j]; temp.x1[j]=data->x1[j]; temp.x2[j]=data->x2[j];
    temp.w[j]=data->w[j];
    for(i=0; i<data->voiNr; i++) {
      temp.voi[i].y[j]=data->voi[i].y[j];
      temp.voi[i].y2[j]=data->voi[i].y2[j];
      temp.voi[i].y3[j]=data->voi[i].y3[j];
    }
  }
  for(i=0; i<data->voiNr; i++) {
    strcpy(temp.voi[i].name, data->voi[i].name);
    strcpy(temp.voi[i].voiname, data->voi[i].voiname);
    strcpy(temp.voi[i].hemisphere, data->voi[i].hemisphere);
    strcpy(temp.voi[i].place, data->voi[i].place);
    temp.voi[i].size=data->voi[i].size;
    temp.voi[i].sw=data->voi[i].sw;
    temp.voi[i].sw2=data->voi[i].sw2;
    temp.voi[i].sw3=data->voi[i].sw3;
  }

  /* Reallocate memory for data */
  dftEmpty(data);
  if(dftSetmem(data, temp.frameNr+1, temp.voiNr)) {dftEmpty(&temp); return 2;}

  /* Set nullframe */
  data->x[0]=data->x1[0]=data->x2[0]=0.0; data->w[0]=0.0;
  for(i=0; i<temp.voiNr; i++)
    data->voi[i].y[0]=data->voi[i].y2[0]=data->voi[i].y3[0]=0.0;

  /* Copy data back from temp */
  data->voiNr=temp.voiNr;
  strcpy(data->unit, temp.unit);
  data->timeunit=temp.timeunit; data->timetype=temp.timetype;
  data->isweight=temp.isweight; data->_type=temp._type;
  strcpy(data->comments, temp.comments);
  for(j=0, n=1; j<temp.frameNr; j++) {
    if(temp.x[j]<0.0) continue;
    if(n==1) data->x2[0]=temp.x1[j];
    data->x[n]=temp.x[j]; data->x1[n]=temp.x1[j]; data->x2[n]=temp.x2[j];
    data->w[n]=temp.w[j];
    for(i=0; i<temp.voiNr; i++) {
      data->voi[i].y[n]=temp.voi[i].y[j];
      data->voi[i].y2[n]=temp.voi[i].y2[j];
      data->voi[i].y3[n]=temp.voi[i].y3[j];
    }
    n++;
  }
  data->frameNr=n;
  for(i=0; i<temp.voiNr; i++) {
    strcpy(data->voi[i].name, temp.voi[i].name);
    strcpy(data->voi[i].voiname, temp.voi[i].voiname);
    strcpy(data->voi[i].hemisphere, temp.voi[i].hemisphere);
    strcpy(data->voi[i].place, temp.voi[i].place);
    data->voi[i].size=temp.voi[i].size;
    data->voi[i].sw=temp.voi[i].sw;
    data->voi[i].sw2=temp.voi[i].sw2;
    data->voi[i].sw3=temp.voi[i].sw3;
  }

  dftEmpty(&temp);

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort DFT regions in alphabetical order by their name. */
int dftSort(DFT *data)
{
  if(data==NULL) return(1);
  if(data->voiNr<=1) return(0);
  qsort(data->voi, data->voiNr, sizeof(Voi), dftQSortName);
  return(0);
}
int dftQSortName(const void *voi1, const void *voi2)
{
  int res;

  res=strcasecmp( ((Voi*)voi1)->name, ((Voi*)voi2)->name );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->voiname, ((Voi*)voi2)->voiname );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->hemisphere, ((Voi*)voi2)->hemisphere );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->place, ((Voi*)voi2)->place );
  return(res);
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort DFT regions in alphabetical order by their plane. */
int dftSortPlane(DFT *data)
{
  if(data==NULL) return(1);
  if(data->voiNr<=1) return(0);
  qsort(data->voi, data->voiNr, sizeof(Voi), dftQSortPlane);
  return(0);
}
int dftQSortPlane(const void *voi1, const void *voi2)
{
  int res;

  res=strcasecmp( ((Voi*)voi1)->place, ((Voi*)voi2)->place );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->name, ((Voi*)voi2)->name );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->voiname, ((Voi*)voi2)->voiname );
  if(res!=0) return(res);
  res=strcasecmp( ((Voi*)voi1)->hemisphere, ((Voi*)voi2)->hemisphere );
  return(res);
}
/*****************************************************************************/

/*****************************************************************************/
/** Check DFT for NA's in sample times and values.
\return Returns the number of NA's that were found.
 */
int dft_nr_of_NA(DFT *dft) {
  int ri, fi, na_nr=0;
  for(fi=0; fi<dft->frameNr; fi++) {
    if(dft->timetype==3) {
      if(dft->x1[fi]==NA) na_nr++;
      if(dft->x2[fi]==NA) na_nr++;
    } else {
      if(dft->x[fi]==NA) na_nr++;
    }
    for(ri=0; ri<dft->voiNr; ri++) if(dft->voi[ri].y[fi]==NA) na_nr++;
  }
  return(na_nr);
}
/*****************************************************************************/

/*****************************************************************************/
/** Replace NA's in basic DFT data with interpolated values.
    If extrapolation is necessary, then the values (0,0) and
    (Infinity,last measured) are assumed.
\return Returns 0, if NA's could be filled with sensible values.
 */
int dftNAfill(DFT *dft)
{
  int ri, fi, fj;
  double x1, x2, y1, y2, x, y;

  if(dft->voiNr<1 || dft->frameNr<1) return(1);
  for(ri=0; ri<dft->voiNr; ri++) for(fi=0; fi<dft->frameNr; fi++) {
    if(dft->x[fi]==NA) return(2);
    if(dft->voi[ri].y[fi]==NA) {
      /* NA's before zero time are always replaced with 0 */
      if(dft->x[fi]<0.0) {dft->voi[ri].y[fi]=0.0; continue;}
      x=dft->x[fi];
      /* Get the previous data that is not NA */
      for(x1=y1=NA, fj=fi-1; fj>=0; fj--) if(dft->voi[ri].y[fj]!=NA) {
        x1=dft->x[fj]; y1=dft->voi[ri].y[fj]; break;
      }
      if(x1==NA || y1==NA) x1=y1=0.0;
      /* Get the following data that is not NA */
      for(x2=y2=NA, fj=fi+1; fj<dft->frameNr; fj++) if(dft->voi[ri].y[fj]!=NA) {
        x2=dft->x[fj]; y2=dft->voi[ri].y[fj]; break;
      }
      if(x2==NA || y2==NA) for(fj=fi-1; fj>=0; fj--) if(dft->voi[ri].y[fj]!=NA) {
        x2=dft->x[fj]; y2=dft->voi[ri].y[fj]; break;
      }
      if(x2==NA || y2==NA) return(2);
      /* Calculate new value */
      if(x2==x1) y=0.5*(y1+y2); else y=y2-(x2-x)*(y2-y1)/(x2-x1);
      dft->voi[ri].y[fi]=y;
    }
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Search the min and max values of DFT TAC data. Data may contain NA's.
\return Returns 0 if successful.
 */
int dftMinMax(
  /** Pointer to the DFT TAC data to search */
  DFT *dft,
  /** Pointer to min X; set to NULL if not needed */
  double *minx,
  /** Pointer to max X; set to NULL if not needed */
  double *maxx,
  /** Pointer to min Y; set to NULL if not needed */
  double *miny,
  /** Pointer to max Y; set to NULL if not needed */
  double *maxy
) {
  int ri, fi;
  double x1, x2, y1, y2;

  if(dft==NULL) return(1);
  x1=x2=y1=y2=NA;
  for(fi=0; fi<dft->frameNr; fi++) {
    if(dft->timetype==3) {
      if(dft->x1[fi]!=NA) {
        if(x1==NA || x1>dft->x1[fi]) x1=dft->x1[fi];
      }
      if(dft->x2[fi]!=NA) {
        if(x2==NA || x2<dft->x2[fi]) x2=dft->x2[fi];
      }
    } else if(dft->x[fi]!=NA) {
      if(x1==NA || x1>dft->x[fi]) x1=dft->x[fi];
      if(x2==NA || x2<dft->x[fi]) x2=dft->x[fi];
    }
    for(ri=0; ri<dft->voiNr; ri++) if(dft->voi[ri].y[fi]!=NA) {
      if(y1==NA || y1>dft->voi[ri].y[fi]) y1=dft->voi[ri].y[fi];
      if(y2==NA || y2<dft->voi[ri].y[fi]) y2=dft->voi[ri].y[fi];
    }
  }
  if(minx!=NULL) {if(x1==NA) return(3); else *minx=x1;}
  if(maxx!=NULL) {if(x2==NA) return(4); else *maxx=x2;}
  if(miny!=NULL) {if(y1==NA) return(5); else *miny=y1;}
  if(maxy!=NULL) {if(y2==NA) return(6); else *maxy=y2;}
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Returns the lowest activity value in DFT */
double dft_kBqMin(DFT *data) {

  int i, j;
  double min=1e+99;

  for(i=0; i<data->voiNr ;i++) {
    for(j=0; j<data->frameNr; j++) {
      if(data->voi[i].y[j]!=NA && data->voi[i].y[j]<min)
	min=data->voi[i].y[j];
    }
  }
  return (min);
}
/*****************************************************************************/

/*****************************************************************************/
/** Returns the highest activity value in DFT */
double dft_kBqMax(DFT *data) {

  int i, j;
  double max=-1e+99;

  for(i=0; i<data->voiNr ;i++){
    for(j=0; j<data->frameNr; j++){
      if(data->voi[i].y[j]!=NA && data->voi[i].y[j]>max)
	max=data->voi[i].y[j];
    }
  }
  return(max);
}
/*****************************************************************************/

/*****************************************************************************/
/** Sorts TAC frames by increasing sample time.
\return Returns 0 if ok.
 */
int dftSortByFrame(DFT *dft)
{
  int ri, fi, fj;
  double d;

  if(dft==NULL || dft->voiNr<1 || dft->frameNr<1) return(1);
  for(fi=0; fi<dft->frameNr-1; fi++) for(fj=fi+1; fj<dft->frameNr; fj++) {
    if(dft->x[fj]>=dft->x[fi]) continue;
    d=dft->x[fi];  dft->x[fi]=dft->x[fj];   dft->x[fj]=d;
    d=dft->x1[fi]; dft->x1[fi]=dft->x1[fj]; dft->x1[fj]=d;
    d=dft->x2[fi]; dft->x2[fi]=dft->x2[fj]; dft->x2[fj]=d;
    d=dft->w[fi];  dft->w[fi]=dft->w[fj];   dft->w[fj]=d;
    for(ri=0; ri<dft->voiNr; ri++) {
      d=dft->voi[ri].y[fi]; dft->voi[ri].y[fi]=dft->voi[ri].y[fj]; dft->voi[ri].y[fj]=d;
      d=dft->voi[ri].y2[fi]; dft->voi[ri].y2[fi]=dft->voi[ri].y2[fj]; dft->voi[ri].y2[fj]=d;
      d=dft->voi[ri].y3[fi]; dft->voi[ri].y3[fi]=dft->voi[ri].y3[fj]; dft->voi[ri].y3[fj]=d;
    }
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Correct frame start and end times if frames are slightly overlapping or
 *  have small gaps in between.
 *  Data is not corrected if time unit is not known to be min or sec.
\return If overlap is considerable (>1 s), or another error is encountered,
    function returns a non-zero value. Otherwise 0 is returned.
 */
int dftDeleteFrameOverlap_old(DFT *dft)
{
  int fi;
  double overlap, overlap_limit=1.8;

  if(dft->timeunit!=TUNIT_MIN && dft->timeunit!=TUNIT_SEC) return(0);
  if(dft->timeunit==TUNIT_MIN) overlap_limit/=60.0;
  for(fi=0; fi<dft->frameNr-1; fi++) {
    overlap=dft->x2[fi] - dft->x1[fi+1];
    if(overlap==0.0) continue;
    else if(fabs(overlap)>overlap_limit) return(2);
    dft->x2[fi]=dft->x1[fi+1];
  }
  return(0);
}
/*****************************************************************************/

/******************************************************************************/
/** Correct frame start and end times if frames are slightly overlapping or
 *  have small gaps in between.
 *  Large gap is not corrected and it does not lead to an error.
\return If overlap is considerable (>1 s), or another error is encountered,
    function returns a non-zero value. Otherwise 0 is returned.
 */
int dftDeleteFrameOverlap(
  /** Pointer to DFT data. Time unit must be set, otherwise no checking is done.
   *  Timetype must be 3, i.e. both frame start and end time must be present. */
  DFT *dft
) {
  int fi;
  double overlap, overlap_limit=1.8, flen1, flen2;

  if(dft==NULL) return(1);
  if(dft->timetype!=3) return(0);
  if(dft->timeunit!=TUNIT_MIN && dft->timeunit!=TUNIT_SEC) return(0);
  if(dft->timeunit==TUNIT_MIN) overlap_limit/=60.0;
  for(fi=0; fi<dft->frameNr-1; fi++) {
    overlap=dft->x2[fi] - dft->x1[fi+1];
    if(overlap==0.0) continue; // no gap or overlap
    else if(overlap<-overlap_limit) continue; // gap is large, then do nothing
    else if(overlap>overlap_limit) return(2); // overlap is large: error
    /* Correct the small gap/overlap by making frame durations more similar */
    flen1=dft->x2[fi]-dft->x1[fi]; flen2=dft->x2[fi+1]-dft->x1[fi+1];
    if(overlap>0.0) { // overlap
      if(flen1>flen2) dft->x2[fi]=dft->x1[fi+1]; else dft->x1[fi+1]=dft->x2[fi];
    } else { // gap
      if(flen1>flen2) dft->x1[fi+1]=dft->x2[fi]; else dft->x2[fi]=dft->x1[fi+1];
    }
  }
  return(0);
}
/*****************************************************************************/

/******************************************************************************/
/** Overwrites DFT comments with information in current DFT header.
 *  If DFT format specifies that titles are to be saved, that
 *  information is not included in comments. */ 
void dftSetComments(DFT *dft)
{
  char tmp[256];

  strcpy(dft->comments, "");
  /* Write in comments the information that will not be included in titles */
  if(dft->scanStartTime[0]) {
    sprintf(tmp, "# scan_start_time := %s\n", dft->scanStartTime);
    strcat(dft->comments, tmp);
  }
  if(dft->injectionTime[0]) {
    sprintf(tmp, "# injection_time := %s\n", dft->injectionTime);
    strcat(dft->comments, tmp);
  }
  strcpy(tmp, "# decay_correction := ");
  if(dft->decayCorrected==1) strcat(tmp, "Yes");
  else if(dft->decayCorrected==2) strcat(tmp, "No");
  else strcat(tmp, "Unknown");
  strcat(dft->comments, tmp);
  if(dft->isotope[0]) {
    sprintf(tmp, "# isotope := %s\n", dft->isotope);
    strcat(dft->comments, tmp);
  }
  if(dft->radiopharmaceutical[0]) {
    sprintf(tmp, "# radiopharmaceutical := %s\n", dft->radiopharmaceutical);
    strcat(dft->comments, tmp);
  }
  /* If titles are set to be saved, then there's no need to put more in comments */
  if(dft->_type==1) return;
  
  /* Ok then, lets write even title information in comments */
  if(dft->studynr[0]) {
    sprintf(tmp, "# study_number := %s\n", dft->studynr);
    strcat(dft->comments, tmp);
  }
  sprintf(tmp, "# timeunit := %s\n", petTunit(dft->timeunit) );
  sprintf(tmp, "# unit := %s\n", dft->unit );
  // Region names and volumes are not saved in comments because of space limit
  
  return;
}
/*****************************************************************************/

/*****************************************************************************/

