/******************************************************************************
  Copyright (c) 2007 by Turku PET Centre

  File:        dftio.c
  Description: Contains I/O functions for DFT files.
  Known bugs:  Line length cannot be larger than max int value, but this is
               not taken into consideration. This may limit the curve number
	       in some systems.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi

  Modification history:
  2007-02-10 Vesa Oikonen
    Contents separated from dft.c.
    Uses new unit enums.
  2007-04-20 VO
    Corrected a 'feature' in saving time unit 'unknown'.
  2007-04-24 VO
    Corrected bug in dftWrite():
    crashed when writing DFT file was not successful.
  2007-08-09 VO
    dftWrite() defaults to 'Times' instead of 'Distances'.
  2007-08-12 VO
    Backup file extension is taken from defined BACKUP_EXTENSION.
  2007-08-13 VO
    Uses new unit enums and functions.


******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
/*****************************************************************************/
#include "decpoint.h"
/*****************************************************************************/
#include "include/dftio.h"
#include "include/dftunit.h"
#include "include/ncifile.h"
#include "include/idwc.h"
#include "include/if.h"
/*****************************************************************************/

/*****************************************************************************/
/** Read DFT file contents into specified data structure, emptying its old
    contents.
\return Returns 0 when successful, in case of error sets dfterrmsg.
 */
int dftRead(char *filename, DFT *data)
{
  FILE *fp;
  char *cptr, *line, *lptr, temp[128];
  int i, j, c, type=0, voiNr=0, frameNr=0, longest=0;
  double f;
  IFT ift;


  /* Empty data */
  dftEmpty(data);

  /* Check from filename if this is an IDWC file */
  if(fncasematch("*.idwc", filename)==1 || fncasematch("*.idw", filename)==1) {
    if(idwcRead(filename, data)==0) return(0); else return(1);
  }
  /* Check from filename if this is an IF file */
  if(fncasematch("*.if", filename)==1) {
    if(ifRead(filename, data)==0) return(0); else return(1);
  }

  /* Try to read information from an interfile-type header */
  iftInit(&ift);
  if(iftRead(&ift, filename, 1)!=0) iftEmpty(&ift);

  /* Open file */
  fp=fopen(filename, "r");
  if(fp==NULL) {
    strcpy(dfterrmsg, "cannot open file"); iftEmpty(&ift); return 1;
  }

  /* Decide file type */
  type=dftType(fp);
  if(type==0) { /* unknown format */
    fclose(fp); iftEmpty(&ift);
    strcpy(dfterrmsg, "unknown file format"); return(1);
  } else if(type==3) {
    strcpy(dfterrmsg, "cannot read fit file"); fclose(fp); iftEmpty(&ift); return 1;
  } else if(type==4) {
    /* Try to read the old format */
    fclose(fp); i=roikbqRead(filename, data); if(i) {iftEmpty(&ift); return 1;}
    dftFrametimes(data); return(0);
  }

  /* Get the length of the longest line */
  i=0; while((c=fgetc(fp))!=EOF) {
    if(c==10 || c==13) {if(i>longest) longest=i; i=0;} else i++;}
  if(i>longest) longest=i;
  rewind(fp); longest+=2;
  /* and allocate memory for string of that length */
  line=(char*)malloc((longest+1)*sizeof(char));
  if(line==NULL) {strcpy(dfterrmsg, "out of memory"); iftEmpty(&ift); fclose(fp); return 2;}

  /* Get the frame number */
  i=0; while(fgets(line, longest, fp)!=NULL && *line) {
    if(line[0]=='#') continue;
    cptr=strchr(line, '#'); if(cptr!=NULL) *cptr=(char)0;
    for(j=0; j<strlen(line); j++)
      if(isalnum((int)line[j]) || line[j]=='.') {i++; break;}
  }
  rewind(fp); frameNr=i; if(type==1) frameNr-=4; /* title lines */
  if(frameNr<1) {strcpy(dfterrmsg, "contains no data"); iftEmpty(&ift); fclose(fp); return 1;}

  /* Get the number of curves */
  /* find first line that is not an empty or a comment line */
  /* In plain data file that is first frame, which starts with time */
  /* In normal DFT file that is 1st title line, which starts with 'DFT' */
  while(fgets(line, longest, fp)!=NULL) {
    lptr=line; if(line[0]=='#') continue;
    cptr=strtok(lptr, " \t\n\r"); if(cptr==NULL) continue;
    i=0; while((cptr=strtok(NULL, " \t\n\r"))!=NULL) i++;
    break;
  }
  rewind(fp); voiNr=i;
  if(voiNr<1) {strcpy(dfterrmsg, "contains no curves"); iftEmpty(&ift); fclose(fp); return 1;}

  /* Allocate memory for data */
  if(dftSetmem(data, frameNr, voiNr)) {
    strcpy(dfterrmsg, "out of memory"); iftEmpty(&ift); fclose(fp); return 2;}

  /* For plain data files, set defaults for title data */
  if(type==2) {
    for(i=0; i<voiNr; i++) {
      sprintf(data->voi[i].voiname, "%06d", i+1);
      strcpy(data->voi[i].name, data->voi[i].voiname);
    }
    /* Default: times in minutes, and frame mid time */
    data->timeunit=TUNIT_UNKNOWN; data->timetype=0;
  }

  /*
   *  Try to read information from an interfile-type header
   */
  if(ift.keyNr>0) dft_fill_hdr_from_IFT(data, &ift);
  iftEmpty(&ift);

  /* Read title lines, if they exist */
  i=0; if(type==1) do {
    if(fgets(line, longest, fp)==NULL) {
      strcpy(dfterrmsg, "wrong format"); iftEmpty(&ift); fclose(fp); return 3;}
    lptr=line;
    /* Check for comment line */
    if(line[0]=='#') { /* Read comment only if there is space left */
      if(strlen(data->comments)+strlen(line)<_DFT_COMMENT_LEN)
        strcat(data->comments, line);
      continue;
    }
    cptr=strchr(line, '#'); if(cptr!=NULL) *cptr=(char)0;
    /* Read first token, and check for empty lines as well */
    cptr=strtok(lptr, " \t\n\r"); if(cptr==NULL) continue; else i++;
    if(i==1) { /* VOI names */
      for(j=0; j<voiNr; j++) {
        if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
          strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
        if(strlen(cptr)>1 || *cptr!='.') strncpy(data->voi[j].voiname, cptr, 6);
        else sprintf(data->voi[j].voiname, "%06d", j);
        strcpy(data->voi[j].name, data->voi[j].voiname);
      }
    } else if(i==2) { /* 2nd VOI names (hemispheres) */
      strncpy(data->studynr, cptr, MAX_STUDYNR_LEN);
      for(j=0; j<voiNr; j++) {
        if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
          strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
        if(strlen(cptr)>1||*cptr!='.') {
          strncpy(data->voi[j].hemisphere,cptr,6);
          strcat(data->voi[j].name, " ");
          strcat(data->voi[j].name, data->voi[j].hemisphere);
        } else {
          strcat(data->voi[j].name, " .");
        }
      }
    } else if(i==3) { /* unit and VOI place (planes) */
      strncpy(data->unit, cptr, 12);
      for(j=0; j<voiNr; j++) {
        if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
          strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
        if(strlen(cptr)>1 || *cptr!='.') {
          strncpy(data->voi[j].place, cptr, 6);
          strcat(data->voi[j].name, " ");
          strcat(data->voi[j].name, data->voi[j].place);
        } else {
          strcat(data->voi[j].name, " .");
        }
      }
    } else if(i==4) { /* time type, time unit, and VOI sizes */
      /* time type */
      if(strcasecmp(cptr, "Time")==0) data->timetype=0;
      else if(strcasecmp(cptr, "Times")==0) data->timetype=3;
      else if(strcasecmp(cptr, "Start")==0) data->timetype=1;
      else if(strcasecmp(cptr, "End")==0) data->timetype=2;
      else if(strcasecmp(cptr, "Distance")==0) data->timetype=0;
      else if(strcasecmp(cptr, "Distances")==3) data->timetype=0;
      else {strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
      /* time unit */
      if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
        strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
      strcpy(temp, ""); j=sscanf(cptr, "(%127s)", temp);
      data->timeunit=petTunitId(temp);
      if(data->timeunit<0) {strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
      /* volumes */
      for(j=0; j<voiNr; j++) {
        if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
          strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
        if(strlen(cptr)==1 && *cptr=='.') data->voi[j].size=0.0;
        else data->voi[j].size=atof_dpi(cptr);
      }
    }
  } while(i<4);


  /* Read data lines */
  i=0; while(fgets(line, longest, fp)!=NULL) {
    lptr=line;
    /* Check for comment line */
    if(line[0]=='#') { /* Read comment only if there is space left */
      if(strlen(data->comments)+strlen(line)<_DFT_COMMENT_LEN)
        strcat(data->comments, line);
      continue;
    }
    cptr=strchr(line, '#'); if(cptr!=NULL) *cptr=(char)0;
    /* Read first token, and check for empty lines as well */
    cptr=strtok(lptr, " \t\n\r"); if(cptr==NULL) continue;
    if(i<frameNr) {
      /* Read time(s) */
      f=atof_dpi(cptr);
      if(data->timetype==3) {
        data->x1[i]=f; cptr=strtok(NULL, " \t\n\r"); data->x2[i]=atof_dpi(cptr);
      } else data->x[i]=f;
      /* curve data */
      for(j=0; j<voiNr; j++) {
        if((cptr=strtok(NULL, " \t\n\r"))==NULL) {
          strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}
        if(strlen(cptr)==1 && *cptr=='.') data->voi[j].y[i]=NA;
        else data->voi[j].y[i]=atof_dpi(cptr);
      }
    }
    i++;
  }
  if(i!=frameNr) {strcpy(dfterrmsg, "wrong format"); fclose(fp); return 3;}

  /* Close file */
  fclose(fp); free((char*)line);

  /* Set voiNr and frameNr, and type */
  data->voiNr=voiNr; data->frameNr=frameNr;
  if(type==2) type=0; data->_type=type;

  /* Set study number from filename, if necessary */
  if(strlen(data->studynr)==0) studynr_from_fname(filename, data->studynr);

  /* If weight 'voi' was read, move it to its own place */
  for(i=0; i<data->voiNr; i++) if(!strcasecmp(data->voi[i].voiname, "weight")) {
    data->isweight=1;
    for(j=0; j<data->frameNr; j++) data->w[j]=data->voi[i].y[j];
    /* Move the following VOIs one step backwards */
    for(c=i+1; c<data->voiNr; c++) if(dftCopyvoi(data, c, c-1)) {
      strcpy(dfterrmsg, "cannot read weight"); return 4;}
    data->voiNr--;
    break;
  }
  /* If no weights were found, set uniform weight */
  if(!data->isweight) for(i=0; i<data->frameNr; i++) data->w[i]=1.0;

  /* Calculate frame mid times, or frame start and end times */
  dftFrametimes(data);

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Determine the type of DFT file.
\return 0=unknown; 1=normal DFT; 2=plain data; 3=fit file ; 4=nci file
 */
int dftType(FILE *fp)
{
  char tmp[256];
  int c;

  /* Find first line that is not empty or comment line */
  rewind(fp);
  while(fgets(tmp, 4, fp) != NULL) {if(strlen(tmp) && tmp[0]!='#') break;}
  rewind(fp);

  /* Check for identification strings */
  if(strncasecmp(tmp, "DFT", 3)==0) return 1;
  else if(strncasecmp(tmp, "FIT1", 3)==0) return 3;
  else if(strncasecmp(tmp, "cpt", 3)==0) return 4;

  /* Binary data? */
  while((c=fgetc(fp))!=EOF) if(!isalnum(c) && !isspace(c) && !isgraph(c)) {
    rewind(fp); return 0;}
  rewind(fp); return 2;
}
/*****************************************************************************/

/*****************************************************************************/
/** Prints to stdout the contents of DFT data structure.
    Mainly for testing purposes.
 */
void dftPrint(DFT *data)
{
  int voi, frame;

  printf("Number of curves: %d     Number of data points: %d\n",
    data->voiNr, data->frameNr);
  printf("Study: '%s'  Unit: '%s'\n", data->studynr, data->unit);
  printf("Time unit and type: %d %d\n", data->timeunit, data->timetype);
  if(strlen(data->radiopharmaceutical))
    printf("Radiopharmaceutical: %s\n", data->radiopharmaceutical);
  if(strlen(data->isotope)) printf("Isotope: %s\n", data->isotope);
  if(strlen(data->scanStartTime))
    printf("Scan start time: %s\n", data->scanStartTime);
  if(strlen(data->injectionTime))
    printf("Injection time: %s\n", data->injectionTime);
  if(data->decayCorrected==1) printf("Corrected for physical decay: yes\n");
  else if(data->decayCorrected==2) printf("Corrected for physical decay: no\n");
  printf("_datasize = %d\n", data->_dataSize);
  for(voi=0; voi<data->voiNr; voi++) {
    if(strlen(data->voi[voi].name)>0)
      printf("\nROI name: '%s' Size: %g\n",
        data->voi[voi].name, data->voi[voi].size);
    else
      printf("\nROI name: '%s' '%s' '%s'  Size: %g\n",
        data->voi[voi].voiname, data->voi[voi].hemisphere, data->voi[voi].place,
        data->voi[voi].size);
    for(frame=0; frame<data->frameNr; frame++) {
      printf("%03d:  %11.3e %11.3e %11.3e    %11.3e %11.3e %11.3e\n",
        frame+1, data->x[frame], data->x1[frame], data->x2[frame],
        data->voi[voi].y[frame], data->voi[voi].y2[frame],
        data->voi[voi].y3[frame]);
    }
  }
  printf("Comments:\n");
  if(strlen(data->comments)>0) printf("%s\n", data->comments);
  printf("Weights:\n");
  if(data->isweight)
    for(frame=0; frame<data->frameNr; frame++)
      printf(" %03d  %11.3e %11.3e  %11.3e\n", frame+1,
        data->x1[frame], data->x2[frame], data->w[frame]);
  else
    printf(" contains no weights.\n");
  return;
}
/*****************************************************************************/

/*****************************************************************************/
/** Write DFT data into specied file.
    If file exists, a backup file (%) is written also.
    The file format specified in data is applied.
 */
int dftWrite(DFT *data, char *filename)
{
  int i, j, n, sw;
  char *cptr, tmp[1024], tmp2[128], is_stdout=0;
  FILE *fp;


  /* Check that there is some data to write */
  if(data->voiNr<1 || data->frameNr<1) {
    strcpy(dfterrmsg, "no data"); return 1;}

  /* If extension is *.HTM(L), write in HTML format */
  cptr=strrchr(filename, '.'); if(cptr!=NULL) cptr++;
  if(cptr!=NULL) {
    if(strcasecmp(cptr, "HTM")==0 || strcasecmp(cptr, "HTML")==0)
      return(dftWriteHTML(data, filename, 1));
  }

  /* Check if writing to stdout */
  if(!strcasecmp(filename, "stdout")) is_stdout=1;

  /* Check if file exists; backup, if necessary */
  if(!is_stdout && access(filename, 0) != -1) {
    strcpy(tmp, filename); strcat(tmp, BACKUP_EXTENSION); if(access(tmp, 0) != -1) remove(tmp);
    n=rename(filename, tmp);
  }

  /* Open output file */
  if(is_stdout) fp=(FILE*)stdout;
  else if((fp = fopen(filename, "w")) == NULL) {
    strcpy(dfterrmsg, "cannot open file"); return 2;}

  /* Write title lines */
  if(data->_type==1) {
    /* 1st line with filetype identification string and region names */
    n=fprintf(fp, "%-20.20s", DFT_VER);
    if(n==0) {strcpy(dfterrmsg, "disk full"); if(!is_stdout) fclose(fp); return 3;}
    for(i=0; i<data->voiNr; i++) fprintf(fp," %-11.11s",data->voi[i].voiname);
    if(data->isweight)  fprintf(fp," %-11.11s", "weight");
    fprintf(fp, "\n");
    /* 2nd line with study identification and 2nd name (hemisphere) */
    if(strlen(data->studynr)) strcpy(tmp, data->studynr); else strcpy(tmp, ".");
    fprintf(fp, "%-20.20s", tmp);
    for(i=0; i<data->voiNr; i++) {
      if(strlen(data->voi[i].hemisphere)) strcpy(tmp, data->voi[i].hemisphere);
      else strcpy(tmp, ".");
      fprintf(fp, " %-11.11s", tmp);
    }
    if(data->isweight) fprintf(fp," %-11.11s", ".");
    fprintf(fp, "\n");
    /* 3rd line with unit and plane names */
    if(strlen(data->unit)) strcpy(tmp, data->unit); else strcpy(tmp, ".");
    fprintf(fp, "%-20.20s", tmp);
    for(i=0; i<data->voiNr; i++) {
      if(strlen(data->voi[i].place)) strcpy(tmp, data->voi[i].place);
      else strcpy(tmp, ".");
      fprintf(fp, " %-11.11s", tmp);
    }
    if(data->isweight) fprintf(fp," %-11.11s", ".");
    fprintf(fp, "\n");
    /* 4th line with time type & unit and region volumes */
    switch(data->timetype) {
      case 0:
        if(data->timeunit==TUNIT_MM || data->timeunit==TUNIT_UM)
	  strcpy(tmp, "Distance ");
        else
	  strcpy(tmp, "Time ");
        break;
      case 1: strcpy(tmp, "Start "); break;
      case 2: strcpy(tmp, "End "); break;
      case 3:
        if(data->timeunit==TUNIT_MM || data->timeunit==TUNIT_UM)
	  strcpy(tmp, "Distances ");
        else
	  strcpy(tmp, "Times ");
        break;
    }
    snprintf(tmp2, 128, "(%s)", petTunit(data->timeunit)); strcat(tmp, tmp2);
    fprintf(fp, "%-20.20s", tmp);
    for(i=0; i<data->voiNr; i++) {
      if(data->voi[i].size>=0.0) sprintf(tmp, "%11.3e", data->voi[i].size);
      else strcpy(tmp, ".");
      fprintf(fp, " %-.11s", tmp);
    }
    if(data->isweight) fprintf(fp," %-.11s", ".");
    fprintf(fp, "\n");
  }

  /* Make sure that only frame mid times are saved without titles */
  if(data->_type==0 && data->timetype>0) sw=0; else sw=data->timetype;

  /* Write data */
  for(j=0; j<data->frameNr; j++) {
    /* Time(s) */
    switch(sw) {
      case 0: fprintf(fp, "%17.5f   ", data->x[j]); break;
      case 1: fprintf(fp, "%17.5f   ", data->x1[j]); break;
      case 2: fprintf(fp, "%17.5f   ", data->x2[j]); break;
      case 3: fprintf(fp, "%10.4f %9.4f", data->x1[j], data->x2[j]); break;
    }
    /* y values */
    for(i=0; i<data->voiNr; i++) {
      if(data->voi[i].y[j]!=NA) sprintf(tmp, "%11.3e", data->voi[i].y[j]);
      else strcpy(tmp, ".");
      fprintf(fp, " %-.11s", tmp);
    }
    if(data->isweight && data->_type==1) {
      sprintf(tmp, "%11.3e", data->w[j]); fprintf(fp," %-.11s", tmp);}
    n=fprintf(fp, "\n");
    if(n==0) {strcpy(dfterrmsg, "disk full"); if(!is_stdout) fclose(fp); return 3;}
  }

  /* Write comments */
  n=strlen(data->comments);
  if(n) for(i=0; i<n; i++) {
    if(i>0 && data->comments[i]=='#' &&
       (data->comments[i-1]!='\n' && data->comments[i-1]!='\r'))
      fputc('\n', fp);
    fputc(data->comments[i], fp);
  }

  /* Close file */
  if(!is_stdout) fclose(fp);

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Write DFT contents in HTML table format
    If file exists, a backup file (+BACKUP_EXTENSION) is written also.
    If "stdout" is given as filename, output is directed to stdout
    In case of an error, description is written in dfterrmsg.
\return Returns 0 if ok.
*/
int dftWriteHTML(
  /** Input DFT */
  DFT *dft,
  /** HTML filename */
  char *fname,
  /** Table orientation: 1=original, 2=transposed */
  int orientation
) {
  int ri, fi, is_stdout=0, ret;
  char tmp[FILENAME_MAX];
  FILE *fp;


  /* Check input */
  strcpy(dfterrmsg, "invalid input to dftWriteHTML()");
  if(dft==NULL || dft->frameNr<1 || dft->voiNr<1) return(1);
  if(fname==NULL || strlen(fname)<1) return(1);
  strcpy(dfterrmsg, "");
  /* Check if writing to stdout */
  if(!strcasecmp(fname, "stdout")) is_stdout=1;

  /* Check if file exists; backup, if necessary */
  if(!is_stdout && access(fname, 0) != -1) {
    strcpy(tmp, fname); strcat(tmp, BACKUP_EXTENSION); rename(fname, tmp);}
  strcpy(dfterrmsg, "cannot write file");

  /* Open output file */
  if(is_stdout) fp=(FILE*)stdout;
  else if((fp=fopen(fname, "w"))==NULL) {
    strcpy(dfterrmsg, "cannot open file"); return(2);}

  /* Write XHTML 1.1 doctype and head */
  if(dftWriteXHTML11_doctype(fp) || dftWriteXHTML11_head(fp, "")) {
    strcpy(dfterrmsg, "disk full"); if(!is_stdout) fclose(fp); return(3);
  }

  /* Start writing the body of the HTML file */
  ret=fprintf(fp, "<body>\n");
  if(ret==0) {strcpy(dfterrmsg, "disk full"); if(!is_stdout) fclose(fp); return(3);}

  /* Start the div for table, and the table */
  fprintf(fp, "\n<div id=\"table\">\n");
  fprintf(fp, "<table>\n");

  /* Write the title lines, if there are titles, into the head of the table */
  if(dft->_type>0) {
    fprintf(fp, "<thead>\n");
    if(orientation!=2) {
      /* Empty cell and Region names */
      fprintf(fp, "<tr><td> </td>\n");
      for(ri=0; ri<dft->voiNr; ri++)
        fprintf(fp, "<th>%s</th>\n", dft->voi[ri].voiname);
      fprintf(fp, "</tr>\n");
      /* Study number and Hemispheres */
      fprintf(fp, "<tr><th>%s</th>\n", dft->studynr);
      for(ri=0; ri<dft->voiNr; ri++)
        fprintf(fp, "<th>%s</th>\n", dft->voi[ri].hemisphere);
      fprintf(fp, "</tr>\n");
      /* Unit and Places */
      fprintf(fp, "<tr><th>%s</th>\n", dft->unit);
      for(ri=0; ri<dft->voiNr; ri++)
        fprintf(fp, "<th>%s</th>\n", dft->voi[ri].place);
      fprintf(fp, "</tr>\n");
      /* Time unit and Volumes */
      fprintf(fp, "<tr><th>%s</th>\n", petTunit(dft->timeunit));
      for(ri=0; ri<dft->voiNr; ri++)
        fprintf(fp, "<th>%g</th>\n", dft->voi[ri].size);
      fprintf(fp, "</tr>\n");
    } else if(orientation==2) {
      /* Study number and Unit */
      fprintf(fp, "<tr><th>%s</th>\n", dft->studynr);
      fprintf(fp, "<th>%s</th></tr>\n", dft->unit);  
    }
    /* End the head of table */
    fprintf(fp, "</thead>\n");
  }

  /* Write the DFT data into the body of the table */
  fprintf(fp, "<tbody>\n");

  /* If transposed, write titles */
  if(orientation==2) {
    fprintf(fp, "<tr><th>Region</th><th>Hemisphere</th><th>Plane</th>\n");
    fprintf(fp, "<th>Volume</th></tr>\n");
  }
  if(orientation!=2) { /* Not transposed */
    for(fi=0; fi<dft->frameNr; fi++) {
      if(fi%2) strcpy(tmp, "evenframe"); else strcpy(tmp, "oddframe");
      /* Time */
      fprintf(fp, "<tr class=\"%s\"><th>%g</th>\n", tmp, dft->x[fi]);
      /* Values */
      for(ri=0; ri<dft->voiNr; ri++)
        fprintf(fp, "<th>%g</th>", dft->voi[ri].y[fi]);
      fprintf(fp, "</tr>\n");
    }
  } else { /* transposed */
    for(ri=0; ri<dft->voiNr; ri++) {
      if(ri%2) strcpy(tmp, "evenframe"); else strcpy(tmp, "oddframe");
      /* Region names and volume */
      fprintf(fp, "<tr class=\"%s\"><th>%s</th><th>%s</th><th>%s</th>\n", tmp,
        dft->voi[ri].voiname, dft->voi[ri].hemisphere, dft->voi[ri].place);
      fprintf(fp, "<td>%g</td>\n", dft->voi[ri].size);
      /* Values */
      for(fi=0; fi<dft->frameNr; fi++) {
        fprintf(fp, "<td>%g</td>", dft->voi[ri].y[fi]);
      }
      fprintf(fp, "</tr>\n");
    }
  }
  /* End the table body and table */
  fprintf(fp, "</tbody></table>\n");

  /* Stop writing the body of the HTML file, and end the file */
  fprintf(fp, "</div>\n");
  ret=fprintf(fp, "</body></html>\n\n");
  if(ret==0) {
    strcpy(dfterrmsg, "disk full");
    if(!is_stdout) fclose(fp); return(3);
  }

  /* Close file */
  if(!is_stdout) fclose(fp);
  strcpy(dfterrmsg, "");

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write XHTML 1.1 doctype into an opened file pointer.
\return Returns 0 if successful, non-zero in case of a failure.
 */
int dftWriteXHTML11_doctype(
  FILE *fp
) {
  int n;
  if(fp==NULL) return(1);
  n=fprintf(fp, "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n");
  if(n<20) return(2);
  n=fprintf(fp, "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n\n");
  if(n<20) return(2);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write XHTML 1.1 head for DFT file into an opened file pointer.
\return Returns 0 if successful, non-zero in case of a failure.
 */
int dftWriteXHTML11_head(
  /** File pointer where to write */
  FILE *fp,
  /** Author name, for example software name */
  char *author_name
) {
  int n;
  if(fp==NULL) return(1);

  n=fprintf(fp, "<head>\n"); if(n<6) return(2);
  fprintf(fp, "  <title>PET data</title>\n");
  fprintf(fp, "  <meta http-equiv=\"content-type\" content=\"text/html; charset=iso-8859-1\" />\n");
  fprintf(fp, "  <meta http-equiv=\"content-language\" content=\"en-gb\" />\n");
  fprintf(fp, "  <meta name=\"author\" content=\"%s\" />\n", author_name);
  fprintf(fp, "  <meta name=\"ProgId\" content=\"Excel.Sheet\" />\n");
  fprintf(fp, "  <link rel=\"icon\" href=\"http://www.turkupetcentre.net/favicon.ico\" type=\"image/x-icon\" />\n");
  fprintf(fp, "  <link rel=\"shortcut icon\" href=\"http://www.turkupetcentre.net/favicon.ico\" type=\"image/x-icon\" />\n");
  /* write local CSS with basic settings in case that external CSS is not available */
  fprintf(fp, "  <style type=\"text/css\">\n");
  fprintf(fp, "    thead {background-color:#999999; color:black;}\n");
  fprintf(fp, "    table {text-align:left; width:100%%; border-collapse:collapse; empty-cells:show;}\n");
  fprintf(fp, "    td {border:1px solid black;}\n");
  fprintf(fp, "    .oddframe {background-color:#FFFFFF; color:black;}\n");
  fprintf(fp, "    .evenframe {background-color:#CCCCCC; color:black;}\n");
  fprintf(fp, "    #regcontainer ul {margin-left:0; padding-left:0;}\n");
  fprintf(fp, "    #regcontainer ul li {display:inline; list-style-type:none;}\n");
  fprintf(fp, "    #regcontainer a {padding:2px 4px;}\n");
  fprintf(fp, "    <!--table\n");
  fprintf(fp, "    	{mso-displayed-decimal-separator:\"\\.\";\n");
  fprintf(fp, "    	mso-displayed-thousand-separator:\" \";}\n");
  fprintf(fp, "    -->\n");
  fprintf(fp, "  </style>\n");
  /* load external CSS with more fancy settings */
  fprintf(fp, "  <link rel=\"stylesheet\" type=\"text/css\" href=\"http://www.turkupetcentre.net/dft.css\" />\n");
  fprintf(fp, "</head>\n");
  if(n<7) return(2);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Read certain keys from IFT and set DFT fields accordingly.
\return Returns the nr of identified keys.
 */
int dft_fill_hdr_from_IFT(
  DFT *dft,
  IFT *ift
) {
  int ki, ri, ok_nr=0, ret;
  char keystr[256], *cptr;


  /* Check for study number */
  strcpy(keystr, "studynr"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "study number"); ki=iftGet(ift, keystr);}
  if(ki==-1) {strcpy(keystr, "study_number"); ki=iftGet(ift, keystr);}
  if(ki>=0) {
    strncpy(dft->studynr, ift->item[ki].value, MAX_STUDYNR_LEN);
    dft->studynr[MAX_STUDYNR_LEN]=(char)0;
    ok_nr++;
  }

  /* Check for time unit */
  strcpy(keystr, "timeunit"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "time unit"); ki=iftGet(ift, keystr);}
  if(ki==-1) {strcpy(keystr, "time_unit"); ki=iftGet(ift, keystr);}
  if(ki==-1) {strcpy(keystr, "Time units"); ki=iftGet(ift, keystr);}
  if(ki>=0) {
    ret=petTunitId(ift->item[ki].value);
    if(ret>=0 && ret!=TUNIT_UNKNOWN) {dft->timeunit=ret; ok_nr++;}
  }

  /* Check for sample unit */
  strcpy(keystr, "unit"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "Activity units"); ki=iftGet(ift, keystr);}
  if(ki>=0) {
    strncpy(dft->unit, ift->item[ki].value, 12); dft->unit[12]=(char)0;
    ok_nr++;
  }

  /* Check for region names */
  ri=0;
  do {
    strcpy(keystr, "voiname"); ki=iftGetNth(ift, keystr, ri+1);
    if(ki>=0) {
      strncpy(dft->voi[ri].name, ift->item[ki].value, MAX_REGIONNAME_LEN);
      dft->voi[ri].name[MAX_REGIONNAME_LEN]=(char)0;
      rnameSplit(ift->item[ki].value, dft->voi[ri].voiname,
        dft->voi[ri].hemisphere, dft->voi[ri].place, 6);
    }
    ri++; ok_nr++;
  } while(ki>=0 && ri<dft->_voidataNr);

  /* Check for region volumes */
  strcpy(keystr, "sizes"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "volumes"); ki=iftGet(ift, keystr);}
  if(ki>=0 && strlen(ift->item[ki].value)) {
    ri=0; cptr=strtok(ift->item[ki].value, " ;\t");
    while(cptr!=NULL && ri<dft->_voidataNr) {
      if(strcmp(cptr, ".")==0) {ri++; continue;}
      dft->voi[ri++].size=atof_dpi(cptr);
      cptr=strtok(NULL, " ;\t");
    }
    ok_nr++;
  }

  /* Check for the name of radiopharmaceutical */
  strcpy(keystr, "radiopharmaceutical"); ki=iftGet(ift, keystr);
  if(ki>=0 && strlen(ift->item[ki].value)) {
    strncpy(dft->radiopharmaceutical, ift->item[ki].value, 31);
    dft->radiopharmaceutical[31]=(char)0;
    ok_nr++;
  }

  /* Check for isotope name */
  strcpy(keystr, "isotope"); ki=iftGet(ift, keystr);
  if(ki>=0 && strlen(ift->item[ki].value)) {
    strncpy(dft->isotope, ift->item[ki].value, 6); dft->isotope[6]=(char)0;
    ok_nr++;
  }

  /* Check if there is anything about correction for physical decay */
  strcpy(keystr, "decay_correction"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "decay correction"); ki=iftGet(ift, keystr);}
  if(ki>=0 && strlen(ift->item[ki].value)) {
    if(strncasecmp(ift->item[ki].value, "Yes", 1)==0) {
      dft->decayCorrected=1; ok_nr++;
    } else if(strncasecmp(ift->item[ki].value, "No", 1)==0) {
      dft->decayCorrected=2; ok_nr++;
    }
  }

  /* Check for injection time */
  strcpy(keystr, "injection time"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "injection_time"); ki=iftGet(ift, keystr);}
  if(ki>=0 && strlen(ift->item[ki].value)) {
    /* check if time is in correct format; otherwise try to correct it */
    if(ift->item[ki].value[4]=='-' && ift->item[ki].value[7]=='-' &&
       ift->item[ki].value[13]==':' && ift->item[ki].value[16]==':') {
      strncpy(dft->injectionTime, ift->item[ki].value, 19);
      dft->injectionTime[19]=(char)0; ok_nr++;
    } else if(ift->item[ki].value[2]=='.' && ift->item[ki].value[5]=='.' &&
              ift->item[ki].value[13]==':' && ift->item[ki].value[16]==':') {
      sprintf(dft->injectionTime, "%4.4s-%2.2s-%2.2s %8.8s",
        ift->item[ki].value+6, ift->item[ki].value+3, ift->item[ki].value,
        ift->item[ki].value+11);
      ok_nr++;
    }
  }

  /* Check for scan start time */
  strcpy(keystr, "scan start time"); ki=iftGet(ift, keystr);
  if(ki==-1) {strcpy(keystr, "scan_start_time"); ki=iftGet(ift, keystr);}
  if(ki==-1) {strcpy(keystr, "scan_start"); ki=iftGet(ift, keystr);}
  if(ki==-1) {strcpy(keystr, "scan start"); ki=iftGet(ift, keystr);}
  if(ki>=0 && strlen(ift->item[ki].value)) {
    /* check if time is in correct format; otherwise try to correct it */
    if(ift->item[ki].value[4]=='-' && ift->item[ki].value[7]=='-' &&
       ift->item[ki].value[13]==':' && ift->item[ki].value[16]==':') {
      strncpy(dft->scanStartTime, ift->item[ki].value, 19);
      dft->scanStartTime[19]=(char)0; ok_nr++;
    } else if(ift->item[ki].value[2]=='.' && ift->item[ki].value[5]=='.' &&
              ift->item[ki].value[13]==':' && ift->item[ki].value[16]==':') {
      sprintf(dft->scanStartTime, "%4.4s-%2.2s-%2.2s %8.8s",
        ift->item[ki].value+6, ift->item[ki].value+3, ift->item[ki].value,
        ift->item[ki].value+11);
      ok_nr++;
    }
  }

  return(ok_nr);
}
/*****************************************************************************/

/*****************************************************************************/
