/******************************************************************************

  Copyright (c) 2007 Turku PET Centre

  Library:      dftunit
  Description:  Functions for setting DFT calibration unit.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi

  Modification history:
  2007-02-10 Vesa Oikonen
    First created.
  2007-08-13 VO
    Units definitions and functions are replaced by the ones in petunits.c.
    
    
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
/*****************************************************************************/
#include "include/dft.h"
#include "include/dftunit.h"
/*****************************************************************************/

/*****************************************************************************/
/** Set DFT calibration unit string */
void dftUnitToDFT(DFT *dft, int dunit)
{
  strcpy(dft->unit, dftUnit(dunit));
}
/*****************************************************************************/

/*****************************************************************************/
/** Conversion of the DFT calibration unit.
 *  Changes both data values and unit string. 
 *  Currently available conversions are:
 *    MBq/cc <-> kBq/cc
 *    kBq/cc <-> Bq/cc
 *    MBq/cc <-> Bq/cc     
 *    nCi/cc <-> kBq/cc
\return Returns 0 if conversion was successful, and <> 0 if failed.
 */   
int dftUnitConversion(
  /** Pointer to existing DFT data, whose calibration unit is to be changed */
  DFT *dft,
  /** New unit code */
  int dunit
) {
  int current_dunit=CUNIT_UNKNOWN;
  double f=1.0;
  int ri, fi;

  /* Check the input */
  if(dft==NULL || dunit<0) return(1);
  
  /* Identify the current unit */
  current_dunit=petCunitId(dft->unit);

  /* Determine the conversion factor to kBq/cc */
  switch(current_dunit) {
    case CUNIT_BQ_PER_ML:  f*=0.001; break;
    case CUNIT_KBQ_PER_ML: f*=1.0; break;
    case CUNIT_MBQ_PER_ML: f*=1000.0; break;
    case CUNIT_NCI_PER_ML: f*=0.037; break;
    default: return(2);
  }

  /* Determine the conversion factor from kBq/cc to the required unit */
  switch(dunit) {
    case CUNIT_BQ_PER_ML:  f*=1000.0; break;
    case CUNIT_KBQ_PER_ML: f*=1.0; break;
    case CUNIT_MBQ_PER_ML: f*=0.001; break;
    case CUNIT_NCI_PER_ML: f/=0.037; break;
    default: return(3);
  }

  /* Convert the data values */
  if(f==1.0) return(0);
  for(fi=0; fi<dft->frameNr; fi++) {
    for(ri=0; ri<dft->voiNr; ri++) {
      if(dft->voi[ri].y[fi]!=NA) dft->voi[ri].y[fi]*=f;
      if(dft->voi[ri].y2[fi]!=NA) dft->voi[ri].y2[fi]*=f;
      if(dft->voi[ri].y3[fi]!=NA) dft->voi[ri].y3[fi]*=f;
    }
  }
  
  /* Set the new unit string */
  dftUnitToDFT(dft, dunit);

  return(0);
} 
/*****************************************************************************/

/*****************************************************************************/
/** Set DFT timeunit; does not change the sample times.
\return Returns 0 if successful, and <> 0 if invalid timeunit.
 */
int dftTimeunitToDFT(DFT *dft, const char *timeunit)
{
  if(dft==NULL || timeunit==NULL) return(1);
  int tunit=petTunitId(timeunit);
  if(tunit<0) return(2); else dft->timeunit=tunit;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Conversion of the DFT timeunit.
 *  Changes both data values and timeunit code. 
 *  Currently available conversions are:
 *    min <-> sec
\return Returns 0 if conversion was successful, and <> 0 if failed.
 */   
int dftTimeunitConversion(
  /** Pointer to existing DFT data, whose timeunit is to be changed */
  DFT *dft,
  /** New timeunit code */
  int tunit
) {
  /* Check the input */
  if(dft==NULL || tunit<0) return(1);
  
  /* Check if unit already is as required */
  if(dft->timeunit==tunit) return(0);
  
  /* Do the conversion, if supported */
  if(dft->timeunit==TUNIT_MIN && tunit==TUNIT_SEC)
    dftMin2sec(dft);
  else if(dft->timeunit==TUNIT_SEC && tunit==TUNIT_MIN)
    dftSec2min(dft);
  else
    return(2);

  return(0);
} 
/*****************************************************************************/

/*****************************************************************************/
/** Change time unit from min to sec, without checking original unit. */
void dftMin2sec(DFT *dft)
{
  int i;

  for(i=0; i<dft->frameNr; i++) {
    if(dft->x[i]!=NA)  dft->x[i]*=60.;
    if(dft->x1[i]!=NA) dft->x1[i]*=60.;
    if(dft->x1[i]!=NA) dft->x2[i]*=60.;
  }
  dft->timeunit=TUNIT_SEC;
}
/*****************************************************************************/

/*****************************************************************************/
/** Change time unit from sec to min, without checking original unit. */
void dftSec2min(DFT *dft)
{
  int i;

  for(i=0; i<dft->frameNr; i++) {
    if(dft->x[i]!=NA)  dft->x[i]/=60.; 
    if(dft->x1[i]!=NA) dft->x1[i]/=60.; 
    if(dft->x1[i]!=NA) dft->x2[i]/=60.;
  }
  dft->timeunit=TUNIT_MIN;
}
/*****************************************************************************/

/*****************************************************************************/
