/******************************************************************************
  Copyright (c) 2002-2007 by Turku PET Centre

  File:        result.c
  Description: Contains functions for reading and writing RES (result) files.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi/

  Modification history:
  2002-05-26 Vesa Oikonen
      First created.
  2002-07-23 VO
      Small changes in printing fields.
  2002-07-27 VO
      resWriteHTML() added.
      resWrite() applies resWriteHTML() if filename extension is *.htm(l)
  2002-07-30 VO
      Switches included in result data structure in result.h.
      Name of study (studynr) included in data structure. Filled in resRead().
      Included function resFName2study().
      Included functions resMedian() and resMean().
  2002-08-14 VO
      Floats -> Doubles.
  2002-10-10 VO
      resSortByName() and resCopyMHeader() added.
  2002-10-17 VO
      One digit less written for negative numbers.
  2002-11-13 VO
      Included function resDelete().
  2003-01-18 VO
      Added CL's and SD's for each result parameter.
  2003-02-18 VO
      Floating point numbers are written with uppercase E in resWriteHTML().
  2003-04-16 VO
      resRead can now read Plasmafile and refroi fields that contain spaces.
  2003-10-13 VO
      resMean() works now also if variance=0.
  2003-12-02 VO
      Added fields for tissue density, lumped constant, and plasma concentration.
  2004-01-26 VO
      Added field for beta.
  2004-07-12 VO
      Added field for Vb.
      Changes in comment style.
  2004-08-17 VO
      Added sw2.
  2004-08-23 VO
      resFName2study() replaced by new library function studynr_from_fname.
  2004-10-13 VO
      tm_isdst=-1 (unknown Daylight saving time).
  2005-10-30 VO
      resInit(): Vb is set to -1.0.
      resWriteHTML(): big changes in HTML codes, but table looks as before.
  2006-06-03 VO
      Added some support for new RES field names.
      resRead(): most key words are now case insensitive.
      resRead(): atof() replaced by decimal point insensitive atof_dpi().
  2006-07-11 VO
      resRead(): reads also units, if they exist.
      resRead(): parameter names are saved in titleline with one space only.
      resPrint(): column widths are determined individually.
  2006-07-31 VO
      resMean(): smaller SD is accepted without setting it to zero.
      Added function resParameterPrintType(), which is used in
      resWrite() and resWriteHTML().
  2007-01-15 VO
      HTML table changed to valid XHTML 1.1.
      Added functions resWriteXHTML11_doctype(), resWriteXHTML11_head(), and
      resIsDuplicateNames().
      Existing backup files are deleted before they are created (renamed).
  2007-08-12 VO
      Backup file extension is taken from defined BACKUP_EXTENSION.
      Argument of strftime() changed to work with current mingw compiler.
  2007-09-13 VO
      Added fields datanr and fitmethod in RES I/O.
      Added function resSelectRegions().
      Return value of localtime() is always checked.

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
/*****************************************************************************/
#include "decpoint.h"
/*****************************************************************************/
#include "include/result.h"
/*****************************************************************************/
/* Local functions */
int resQSortComp(const void *f1, const void *f2);
int resQSortName(const void *voi1, const void *voi2);
/*****************************************************************************/

/*****************************************************************************/
/** Free memory allocated for results. All data are cleared. */
void resEmpty(RES *res)
{
  if(res->_voidataNr>0) {
    free((char*)(res->voi));
    res->_voidataNr=0;
  }
  res->voiNr=0;
  res->parNr=0;
  res->program[0]=(char)0; res->refroi[0]=(char)0; res->datarange[0]=(char)0;
  res->datanr=0; res->fitmethod[0]=(char)0;
  res->datafile[0]=res->reffile[0]=res->plasmafile[0]=res->bloodfile[0]=(char)0;
  res->density=res->lc=res->concentration=res->beta=0.0;
  res->Vb=-1.0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Initiate RES structure. This should be called once before first use. */
void resInit(RES *res)
{
  memset(res, 0, sizeof(RES));
  res->_voidataNr=0; res->voiNr=0; res->parNr=0; res->Vb=-1.0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Allocate memory for result data. Old data is destroyed. */
int resSetmem(RES *res, int voiNr)
{
  int ri, pi;

  /* Check that there is something to do */
  if(voiNr<1) return(1);

  /* Clear previous data, but only if necessary */
  if(res->_voidataNr>0 || res->voiNr>0) resEmpty(res);

  /* Allocate memory for regional curves */
  res->voi=(ResVOI*)calloc(voiNr, sizeof(ResVOI));
  if(res->voi==NULL) return(2);
  res->_voidataNr=voiNr;

  /* Set SDs and CLs to NA */
  for(ri=0; ri<res->_voidataNr; ri++) for(pi=0; pi<MAX_RESPARAMS; pi++)
    res->voi[ri].sd[pi]=res->voi[ri].cl1[pi]=res->voi[ri].cl2[pi]=NA;

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Print to stdout the contents of RES data structure. */
void resPrint(RES *res)
{
  resWrite(res, "stdout");
}
/*****************************************************************************/

/*****************************************************************************/
/** Read RES file contents to the specified data structure,
    emptying its old contents.
\return In case of an error, >0 is returned, and a description is written in reserrmsg.
 */
int resRead(char *filename, RES *res)
{
  FILE *fp;
  char *cptr, line[1024], *lptr, tmp[1024];
  int i, n;
  int yy, mm, dd, h, m, s;
  struct tm *st;
  time_t timet;
  long bookmark;
  fpos_t file_loc;


  if(RESULT_TEST) printf("resRead(%s, *res);\n", filename);
  /* Empty data */
  resEmpty(res);

  /* Open file */
  fp=fopen(filename, "r");
  if(fp==NULL) {strcpy(reserrmsg, "cannot open file"); return(1);}

  /*
   * Read data, each result set separately, saving only the first one
   */
  strcpy(reserrmsg, "wrong format");

  /* Read program name */
  while(fgets(line, 1024, fp)!=NULL) {
    /* Ignore empty and comment lines */
    if(strlen(line)<4 || line[0]=='#') continue; else break;
  }
  /* Check for string (c) or (C) */
  if(strstr(line, "(c)") || strstr(line, "(C)")) {
    strcpy(res->program, line);
  } else {fclose(fp); return(2);}

  /* Read calculation date and time */
  while(fgets(line, 1024, fp)!=NULL) if(strlen(line)>2 && line[0]!='#') break;
  if(strncasecmp(line, "Date:", 5)) {fclose(fp); return(4);}
  lptr=&line[5]; cptr=strtok(lptr, " \t\n\r");
  if(cptr!=NULL) {
    timet=time(NULL); st=localtime(&timet);
    sscanf(cptr, "%d-%d-%d", &yy, &mm, &dd);
    st->tm_mday=dd; st->tm_mon=mm-1; st->tm_year=yy-1900;
    cptr=strtok(NULL, " \t\n\r");
    if(cptr!=NULL) sscanf(cptr, "%d:%d:%d", &h, &m, &s);
    st->tm_hour=h; st->tm_min=m; st->tm_sec=s; st->tm_isdst=-1;
    res->time=mktime(st);
  }

  /* Read studynr, datafiles, ref region, data range, etc */
  do {
    while(fgets(line, 1024, fp)!=NULL) if(strlen(line)>2 && line[0]!='#') break;
    n=0;
    if(strncasecmp(line, "Study", 5)==0) {
      lptr=&line[6]; cptr=strtok(lptr, " \t\n\r"); n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strncpy(res->studynr, cptr, MAX_STUDYNR_LEN);
    } else if(strncasecmp(line, "Data file", 9)==0) {
      lptr=&line[10]; cptr=strtok(lptr, " \t\n\r"); n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strcpy(res->datafile, cptr);
    } else if(strncasecmp(line, "ROI file", 8)==0) {
      lptr=&line[9]; cptr=strtok(lptr, " \t\n\r"); n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strcpy(res->datafile, cptr);
    } else if(strncasecmp(line, "Plasma file", 11)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strcpy(res->plasmafile, cptr);
      cptr=res->plasmafile+strlen(res->plasmafile)-1;
      while(isspace((int)*cptr)) {*cptr=(char)0; cptr--;}
    } else if(strncasecmp(line, "Blood file", 10)==0) {
      lptr=&line[11]; cptr=strtok(lptr, " \t\n\r"); n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strcpy(res->bloodfile, cptr);
    } else if(strncasecmp(line, "Reference file", 14)==0) {
      lptr=&line[15]; cptr=strtok(lptr, " \t\n\r"); n=1;
      if(cptr!=NULL && strlen(cptr)<1024) strcpy(res->reffile, cptr);
    } else if(strncasecmp(line, "Reference region", 16)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL && strlen(cptr)<64) strcpy(res->refroi, cptr);
      cptr=res->refroi+strlen(res->refroi)-1;
      while(isspace((int)*cptr)) {*cptr=(char)0; cptr--;}
    } else if(strncasecmp(line, "Fit time", 8)==0 || strncmp(line, "Data range", 10)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL && strlen(cptr)<128) strcpy(res->datarange, cptr);
      cptr=res->datarange+strlen(res->datarange)-1;
      while(isspace((int)*cptr)) {*cptr=(char)0; cptr--;}
    } else if(strncasecmp(line, "Data nr", 7)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->datanr=atoi(cptr);
    } else if(strncasecmp(line, "Fit method", 10)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL && strlen(cptr)<128) strcpy(res->fitmethod, cptr);
      cptr=res->fitmethod+strlen(res->fitmethod)-1;
      while(isspace((int)*cptr)) {*cptr=(char)0; cptr--;}
    } else if(strncasecmp(line, "Tissue density", 14)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->density=atof_dpi(cptr);
    } else if(strncasecmp(line, "Lumped constant", 15)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->lc=atof_dpi(cptr);
    } else if(strncasecmp(line, "Concentration", 13)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->concentration=atof_dpi(cptr);
    } else if(strncasecmp(line, "Beta", 4)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->beta=atof_dpi(cptr);
    } else if(strncasecmp(line, "Vb", 2)==0) {
      cptr=strchr(line, ':')+1; while(isspace((int)*cptr)) cptr++; n=1;
      if(cptr!=NULL) res->Vb=atof_dpi(cptr);
    }
  } while(n);

  /* Check now if data was weighted or not from the line read above */
  if(strncasecmp(line, "Data was weighted", 17)==0) res->isweight=1;
  else if(strncasecmp(line, "Data was not weighted", 21)==0) res->isweight=0;
  else {fclose(fp); return(6);}

  /* Read the result parameter title line */
  while(fgets(line, 1024, fp)!=NULL) if(strlen(line)>2 && line[0]!='#') break;
  if(strncasecmp(line, "Region  ", 8)) {fclose(fp); return(10);}

  cptr=strtok(line+20, " \n\r\t");
  if(cptr!=NULL) strcpy(res->titleline, cptr); else strcpy(res->titleline, "");
  while(cptr!=NULL) {
    cptr=strtok(NULL, " \n\r\t");
    if(cptr!=NULL) {strcat(res->titleline, " "); strcat(res->titleline, cptr);}
  }

  /* Read the result parameter unit line */
  fgetpos(fp, &file_loc);
  while(fgets(line, 1024, fp)!=NULL) if(strlen(line)>2) break;
  if(strncasecmp(line, "# Units : ", 7)==0 || strncasecmp(line, "Units : ", 7)==0) {
    cptr=strtok(line+20, " \n\r\t");
    if(cptr!=NULL) strcpy(res->unitline, cptr); else strcpy(res->unitline, "");
    while(cptr!=NULL) {
      cptr=strtok(NULL, " \n\r\t");
      if(cptr!=NULL) {strcat(res->unitline, " "); strcat(res->unitline, cptr);}
    }
  } else {
    fsetpos(fp, &file_loc);
  }

  /* Read the nr of result lines */
  n=0; bookmark=ftell(fp); if(bookmark<0) {fclose(fp); return(21);}
  while(fgets(line, 1024, fp)!=NULL) {
    if(strlen(line)<20 && line[0]!='#') break;
    n++;
  }
  if(fseek(fp, bookmark, SEEK_SET)) {fclose(fp); return(22);}
  if(RESULT_TEST>1) printf("nr of result lines is %d\n", n);

  /* Allocate memory for regional results */
  if(resSetmem(res, n)) {
    strcpy(reserrmsg, "cannot allocate memory");
    fclose(fp); return(25);
  }

  /* Read regional results */
  res->voiNr=0;
  while(fgets(line, 1024, fp)!=NULL) {
    if(line[0]=='#') continue;
    if(strlen(line)<20) break;
    strcpy(tmp, line);
    /* Read region names */
    lptr=line;
    cptr=strtok(lptr, " \t\n\r"); if(cptr==NULL) {fclose(fp); return(31);}
    strncpy(res->voi[res->voiNr].name, cptr, MAX_REGIONNAME_LEN);
    strncpy(res->voi[res->voiNr].voiname, cptr, 6);
    cptr=strtok(NULL, " \t\n\r"); if(cptr==NULL) {fclose(fp); return(31);}
    if(*cptr!='.') {
      strncpy(res->voi[res->voiNr].hemisphere, cptr, 6);
      strcat(res->voi[res->voiNr].name, " ");
      strcat(res->voi[res->voiNr].name, cptr);
    } else {
      res->voi[res->voiNr].hemisphere[0]=(char)0;
      strcat(res->voi[res->voiNr].name, " .");
    }
    cptr=strtok(NULL, " \t\n\r"); if(cptr==NULL) {fclose(fp); return(31);}
    if(*cptr!='.') {
      strncpy(res->voi[res->voiNr].place, cptr, 6);
      strcat(res->voi[res->voiNr].name, " ");
      strcat(res->voi[res->voiNr].name, cptr);
    } else {
      res->voi[res->voiNr].place[0]=(char)0;
      strcat(res->voi[res->voiNr].name, " .");
    }
    /* Read results */
    i=0; lptr=tmp+20; cptr=strtok(lptr, " \t\n\r");
    if(cptr==NULL) {fclose(fp); return(32);}
    while(cptr!=NULL && i<MAX_RESPARAMS) {
      if(strlen(cptr)==1 && *cptr=='.') res->voi[res->voiNr].parameter[i++]=NA;
      else res->voi[res->voiNr].parameter[i++]=atof_dpi(cptr);
      cptr=strtok(NULL, " \t\n\r");
    }
    if(res->voiNr==0) {res->parNr=i;} else if(i<res->parNr) res->parNr=i;
    /* If 'region' name implies that this was confidence limit or sd, then */
    /* move the values into correct places, and do not increase the voiNr */
    if(res->voiNr==0) {res->voiNr++; continue;}
    if(strcasecmp(res->voi[res->voiNr].voiname, "CL")==0) {
      if(strcmp(res->voi[res->voiNr].hemisphere, "95%")==0) {
        if(strcasecmp(res->voi[res->voiNr].place, "Lower")==0)
          for(i=0; i<res->parNr; i++)
            res->voi[res->voiNr-1].cl1[i]=res->voi[res->voiNr].parameter[i];
        else if(strcasecmp(res->voi[res->voiNr].place, "Upper")==0)
          for(i=0; i<res->parNr; i++)
            res->voi[res->voiNr-1].cl2[i]=res->voi[res->voiNr].parameter[i];
        continue;
      }
    } else if(strcasecmp(res->voi[res->voiNr].voiname, "SD")==0) {
      for(i=0; i<res->parNr; i++)
        res->voi[res->voiNr-1].sd[i]=res->voi[res->voiNr].parameter[i];
      continue;
    }
    res->voiNr++;
  }
  if(res->parNr==0) {fclose(fp); return(33);}
  if(RESULT_TEST) printf("nr of results: %d ; nr of parameters: %d\n", res->voiNr, res->parNr);

  /* Seek for other results in the same file */
  while(fgets(line, 1024, fp)!=NULL) {
    /* Ignore empty and comment lines */
    if(strlen(line)<4 || line[0]=='#') continue; else break;
  }
  /* Check again for string (c) or (C) */
  if(strstr(line, "(c)") || strstr(line, "(C)")) {
    fprintf(stderr, "Warning: %s contains more than one set of results; only the 1st one is used.\n", filename);
  }

  /* Close file */
  fclose(fp);
  strcpy(reserrmsg, "");

  /* Fill studynr if it was not found in file */
  if(!res->studynr[0]) studynr_from_fname(filename, res->studynr);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write calculation results into specied file.
  If file exists, a backup file (+BACKUP_EXTENSION) is written also.
  If "stdout" is given as filename, output is directed to stdout.
  If filename extension is *.htm(l), file is saved in HTML format.
\return In case of an error, >0 is returned, and a description is written in reserrmsg.
*/
int resWrite(RES *res, char *filename)
{
  int i, j, n;
  char tmp[1024], is_stdout=0, *cptr;
  struct tm *st;
  FILE *fp;
  int partype[MAX_RESPARAMS]; /* 0=int, 1=double, 2=exp */
  int colwidth[MAX_RESPARAMS];
  double *p;


  if(RESULT_TEST) printf("resWrite(*res, %s)\n", filename);
  /* Check that there is some data to write */
  if(res==NULL) {strcpy(reserrmsg, "error in result data"); return(1);}
  if(res->voiNr<1) {strcpy(reserrmsg, "no result data"); return(1);}

  /* Write results in HTML format, if necessary */
  cptr=strrchr(filename, '.');
  if(cptr!=NULL && (!strncmp(cptr, ".htm", 4) || !strncmp(cptr, ".HTM", 4)))
    return(resWriteHTML(res, filename));

  /* Check if writing to stdout */
  if(!strcasecmp(filename, "stdout")) is_stdout=1;

  /* Check if file exists; backup, if necessary */
  if(!is_stdout && access(filename, 0) != -1) {
    strcpy(tmp, filename); strcat(tmp, BACKUP_EXTENSION);
    if(access(tmp, 0)!=-1) remove(tmp);
    rename(filename, tmp);
  }
  strcpy(reserrmsg, "cannot write file");

  /* Open output file */
  if(is_stdout) fp=(FILE*)stdout;
  else if((fp = fopen(filename, "w")) == NULL) {
    strcpy(reserrmsg, "cannot open file"); return(2);}

  /* Program name */
  n=fprintf(fp, "%s\n", res->program);
  if(n==0) {
    strcpy(reserrmsg, "disk full");
    if(!is_stdout) fclose(fp); return(3);
  }

  /* Write calculation date and time */
  st=localtime(&res->time); 
  if(st!=NULL) strftime(tmp, 256, "%Y-%m-%d %H:%M:%S", st);
  else strcpy(tmp, "1900-01-01 00:00:00");
  fprintf(fp, "Date:         %s\n", tmp);

  /* Write the studynr */
  if(res->studynr[0]) fprintf(fp, "Study:        %s\n", res->studynr);

  /* Write the names of the original datafiles */
  if(res->datafile[0])   fprintf(fp, "Data file:    %s\n", res->datafile);
  if(res->plasmafile[0]) fprintf(fp, "Plasma file:  %s\n", res->plasmafile);
  if(res->bloodfile[0])  fprintf(fp, "Blood file:   %s\n", res->bloodfile);
  if(res->reffile[0])    fprintf(fp, "Reference file:  %s\n", res->reffile);
  if(res->refroi[0])     fprintf(fp, "Reference region:  %s\n", res->refroi);

  /* Write data range etc */
  if(res->datarange[0])  fprintf(fp, "Data range:   %s\n", res->datarange);
  if(res->datanr>0)      fprintf(fp, "Data nr:      %d\n", res->datanr);
  if(res->fitmethod[0])  fprintf(fp, "Fit method:   %s\n", res->fitmethod);

  /* Write constants */
  if(res->density>0.0)   fprintf(fp, "Tissue density:  %g\n", res->density);
  if(res->lc>0.0)        fprintf(fp, "Lumped constant: %g\n", res->lc);
  if(res->concentration>0.0)
                         fprintf(fp, "Concentration:   %g\n", res->concentration);
  if(res->beta>0.0)      fprintf(fp, "Beta:         %g\n", res->beta);
  if(res->Vb>=0.0)       fprintf(fp, "Vb:           %g %%\n", res->Vb);

  /* Weighting */
  if(res->isweight) fprintf(fp, "Data was weighted.\n");
  else fprintf(fp, "Data was not weighted.\n");

  /* Determine column widths: */
  for(j=0; j<res->parNr; j++) colwidth[j]=1;
  /* should column be printed as integers (0), floats (1) or exponentials (2)? */
  for(j=0; j<res->parNr; j++) {
    partype[j]=resParameterPrintType(res, j);
  }
  /* min width required by titles */
  strcpy(tmp, res->titleline); cptr=strtok(tmp, " \t\n\r");
  j=0; while(cptr!=NULL && j<res->parNr) {
    colwidth[j]=strlen(cptr); cptr=strtok(NULL, " \t\n\r"); j++;
  }
  if(RESULT_TEST>2) {
    printf("col widths after titles were checked:\n");
    for(i=0; i<res->parNr; i++)
      printf("  par%d : partype=%d colwidth=%d\n", i+1, partype[i], colwidth[i]);
  }
  /* min width required by units */
  if(res->unitline[0]) {
    strcpy(tmp, res->unitline); cptr=strtok(tmp, " \t\n\r");
    j=0; while(cptr!=NULL && j<res->parNr) {
      n=strlen(cptr); if(n>colwidth[j]) colwidth[j]=n;
      cptr=strtok(NULL, " \t\n\r"); j++;
    }
    if(RESULT_TEST>2) {
      printf("col widths after units were checked:\n");
      for(i=0; i<res->parNr; i++)
        printf("  par%d : partype=%d colwidth=%d\n", i+1, partype[i], colwidth[i]);
    }
  }
  /* widths required by result values */
  for(i=0; i<res->voiNr; i++) {
    p=res->voi[i].parameter;
    for(j=0; j<res->parNr; j++) {
      if(p[j]==NA) continue;
      if(p[j]>=0) n=4; else n=3;
      switch(partype[j]) {
        case 0: sprintf(tmp, "%.0f", p[j]); break;
        case 1: sprintf(tmp, "%.*f", n, p[j]); break;
        case 2:
        default: sprintf(tmp, "%.*e", n, p[j]); break;
      }
      n=strlen(tmp); if(n>colwidth[j]) colwidth[j]=n;
    }
  }
  if(RESULT_TEST>2) {
    printf("col widths after result values were checked:\n");
    for(i=0; i<res->parNr; i++)
      printf("  par%d : partype=%d colwidth=%d\n", i+1, partype[i], colwidth[i]);
  }


  /* Title line */
  strcpy(tmp, res->titleline); cptr=strtok(tmp, " \t\n\r");
  fprintf(fp, "\n%-20.20s ", "Region");
  j=0; while(cptr!=NULL && j<res->parNr) {
    fprintf(fp, " %-*.*s", colwidth[j], colwidth[j], cptr);
    cptr=strtok(NULL, " \t\n\r"); j++;
  }
  fprintf(fp, "\n");

  /* Write units, if they exist, currently as comment line */
  if(res->unitline[0]) {
    strcpy(tmp, res->unitline); cptr=strtok(tmp, " \t\n\r");
    fprintf(fp, "%-20.20s ", "# Units :");
    j=0; while(cptr!=NULL && j<res->parNr) {
      fprintf(fp, " %-*.*s", colwidth[j], colwidth[j], cptr);
      cptr=strtok(NULL, " \t\n\r"); j++;
    }
    fprintf(fp, "\n");
  }

  /* Write regional results */
  for(i=0; i<res->voiNr; i++) {
    if(res->voi[i].voiname[0]) strcpy(tmp, res->voi[i].voiname); else strcpy(tmp, ".");
    fprintf(fp, "%-6.6s ", tmp);
    if(res->voi[i].hemisphere[0]) strcpy(tmp, res->voi[i].hemisphere); else strcpy(tmp, ".");
    fprintf(fp, "%-6.6s ", tmp);
    if(res->voi[i].place[0]) strcpy(tmp, res->voi[i].place); else strcpy(tmp, ".");
    fprintf(fp, "%-6.6s ", tmp);
    p=res->voi[i].parameter;
    for(j=0; j<res->parNr; j++) {
      if(p[j]==NA) {fprintf(fp, " -%*.*s", colwidth[j], colwidth[j], "."); continue;}
      switch(partype[j]) {
        case 0: fprintf(fp, " %*.0f", colwidth[j], p[j]); break;
        case 1:
          if(p[j]>=0) n=4; else n=3;
          fprintf(fp, " %*.*f", colwidth[j], n, p[j]);
          break;
        default:
          if(p[j]>=0) n=4; else n=3;
          fprintf(fp, " %*.*e", colwidth[j], n, p[j]);
        break;
      }
    }
    fprintf(fp, "\n");
    /* Write SD's, if they exist */
    p=res->voi[i].sd; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "%6.6s %6.6s %6.6s ", "SD", ".", ".");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) {
          switch(partype[j]) {
            case 0: fprintf(fp, " %*.0f", colwidth[j], p[j]); break;
            case 1:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*f", colwidth[j], n, p[j]);
              break;
            default:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*e", colwidth[j], n, p[j]);
              break;
          }
        } else {
          fprintf(fp, " %*.*s", colwidth[j], colwidth[j], ".");
        }
      }
      fprintf(fp, "\n");
    }
    /* Write lower confidence limits, if they exist */
    p=res->voi[i].cl1; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "%6.6s %6.6s %6.6s ", "CL", "95%", "Lower");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) {
          switch(partype[j]) {
            case 0: fprintf(fp, " %*.0f", colwidth[j], p[j]); break;
            case 1:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*f", colwidth[j], n, p[j]);
              break;
            default:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*e", colwidth[j], n, p[j]);
              break;
          }
        } else {
          fprintf(fp, " %*.*s", colwidth[j], colwidth[j], ".");
        }
      }
      fprintf(fp, "\n");
    }
    /* Write upper confidence limits, if they exist */
    p=res->voi[i].cl2; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "%6.6s %6.6s %6.6s ", "CL", "95%", "Upper");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) {
          switch(partype[j]) {
            case 0: fprintf(fp, " %*.0f", colwidth[j], p[j]); break;
            case 1:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*f", colwidth[j], n, p[j]);
              break;
            default:
              if(p[j]>=0) n=4; else n=3;
              fprintf(fp, " %*.*e", colwidth[j], n, p[j]);
              break;
          }
        } else {
          fprintf(fp, " %*.*s", colwidth[j], colwidth[j], ".");
        }
      }
      fprintf(fp, "\n");
    }
  } /* next region */

  /* Close file */
  if(!is_stdout) fclose(fp);
  strcpy(reserrmsg, "");

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write calculation results into specied XHTML 1.1 file.
    If file exists, a backup file (+BACKUP_EXTENSION) is written also.
    If "stdout" is given as filename, output is directed to stdout
\return In case of an error, >0 is returned, and a description is written in reserrmsg.
 */
int resWriteHTML(RES *res, char *fname)
{
  int i, j, n, is_stdout=0;
  char tmp[1024], *cptr;
  FILE *fp;
  struct tm *st;
  int partype[MAX_RESPARAMS]; /* 0=int, 1=double, 2=exp */
  double *p;


  if(RESULT_TEST) printf("resWriteHTML(*res, %s)\n", fname);
  /* Check that there is some data to write */
  if(res==NULL) {strcpy(reserrmsg, "error in result data"); return(1);}
  if(res->voiNr<1) {strcpy(reserrmsg, "no result data"); return(1);}
  /* Check if writing to stdout */
  if(!strcasecmp(fname, "stdout")) is_stdout=1;

  /* Check if file exists; backup, if necessary */
  if(!is_stdout && access(fname, 0) != -1) {
    strcpy(tmp, fname); strcat(tmp, BACKUP_EXTENSION);
    if(access(tmp, 0)!=-1) remove(tmp);
    rename(fname, tmp);
  }
  strcpy(reserrmsg, "cannot write file");

  /* Open output file */
  if(is_stdout) fp=(FILE*)stdout;
  else if((fp=fopen(fname, "w"))==NULL) {
    strcpy(reserrmsg, "cannot open file"); return(2);}

  /* Write XHTML 1.1 doctype and head */
  if(resWriteXHTML11_doctype(fp) || resWriteXHTML11_head(fp, res->program)) {
    strcpy(reserrmsg, "disk full");
    if(!is_stdout) fclose(fp); return(3);
  }

  /* Start writing the body of the HTML file */
  fprintf(fp, "\n<body>\n");

  /* Start the div for tables */
  fprintf(fp, "\n<div id=\"tables\">\n");

  /* Write the title lines to the head of table */
  fprintf(fp, "<table>\n");
  fprintf(fp, "<thead>\n");
  /* Program name */
  fprintf(fp, "<tr><th>Program:</th><td>%s</td></tr>\n", res->program);
  /* Write calculation date and time */
  st=localtime(&res->time);
  if(st!=NULL) strftime(tmp, 256, "%Y-%m-%d %H:%M:%S", st);
  else strcpy(tmp, "1900-01-01 00:00:00");
  fprintf(fp, "<tr><th>Date:</th><td>%s</td></tr>\n", tmp);
  /* Write the studynr */
  if(res->studynr[0])
    fprintf(fp, "<tr><th>Study:</th><td>%s</td></tr>\n", res->studynr);
  /* Write the names of the original datafiles */
  if(res->datafile[0])
    fprintf(fp, "<tr><th>Data file:</th><td>%s</td></tr>\n", res->datafile);
  if(res->plasmafile[0])
    fprintf(fp, "<tr><th>Plasma file:</th><td>%s</td></tr>\n", res->plasmafile);
  if(res->bloodfile[0])
    fprintf(fp, "<tr><th>Blood file:</th><td>%s</td></tr>\n", res->bloodfile);
  if(res->reffile[0])
    fprintf(fp, "<tr><th>Reference file:</th><td>%s</td></tr>\n", res->reffile);
  if(res->refroi[0])
    fprintf(fp, "<tr><th>Reference region:</th><td>%s</td></tr>\n", res->refroi);
  /* Write data range etc */
  if(res->datarange[0])
    fprintf(fp, "<tr><th>Data range:</th><td>%s</td></tr>\n", res->datarange);
  if(res->datanr>0)
    fprintf(fp, "<tr><th>Data nr:</th><td>%d</td></tr>\n", res->datanr);
  if(res->fitmethod[0])
    fprintf(fp, "<tr><th>Fit method:</th><td>%s</td></tr>\n", res->fitmethod);
  /* Write constants */
  if(res->density>0.0)
    fprintf(fp, "<tr><th>Tissue density:</th><td>%g</td></tr>\n", res->density);
  if(res->lc>0.0)
    fprintf(fp, "<tr><th>Lumped constant:</th><td>%g</td></tr>\n", res->lc);
  if(res->concentration>0.0)
    fprintf(fp, "<tr><th>Concentration:</th><td>%g</td></tr>\n", res->concentration);
  if(res->beta>0.0)
    fprintf(fp, "<tr><th>Beta:</th><td>%g</td></tr>\n", res->beta);
  if(res->Vb>=0.0)
    fprintf(fp, "<tr><th>Vb:</th><td>%g %%</td></tr>\n", res->Vb);
  /* Weighting */
  if(res->isweight)
    fprintf(fp, "<tr><th>Weighting:</th><td>yes</td></tr>\n");
  else
    fprintf(fp, "<tr><th>Weighting:</th><td>none</td></tr>\n");
  /* End the head of table */
  fprintf(fp, "</thead>\n");


  /* Collect info on column types */
  for(j=0; j<res->parNr; j++) {
    /* should column be printed as integers, floats or exponentials? */
    partype[j]=resParameterPrintType(res, j);
  }

  /* Write the result data to the table body */
  fprintf(fp, "<tbody>\n");
  /* Write the result title line */
  fprintf(fp, "<tr align=left><th>Region</th><th>Hemisphere</th><th>Plane</th>\n");
  strcpy(tmp, res->titleline); cptr=strtok(tmp, " \t\n\r");
  i=0; while(cptr!=NULL && i<res->parNr) {
    fprintf(fp, "<th>%s</th>", cptr); cptr=strtok(NULL, " \t\n\r"); i++;}
  fprintf(fp, "</tr>\n");
  /* Write regional results */
  for(i=0; i<res->voiNr; i++) {
    if(i%2) strcpy(tmp, "evenroi"); else strcpy(tmp, "oddroi");
    fprintf(fp, "<tr class=\"%s\">", tmp);
    fprintf(fp, "<th>%s</th>", res->voi[i].voiname);
    fprintf(fp, "<th>%s</th>", res->voi[i].hemisphere);
    fprintf(fp, "<th>%s</th>", res->voi[i].place);
    p=res->voi[i].parameter;
    for(j=0; j<res->parNr; j++) switch(partype[j]) {
      case 0: fprintf(fp, "<th>%.0f</th>", p[j]); break;
      case 1: fprintf(fp, "<th>%.4f</th>", p[j]); break;
      default: fprintf(fp, "<th>%.4E</th>", p[j]); break;
    }
    fprintf(fp, "</tr>\n");
    /* Write SD's, if they exist */
    p=res->voi[i].sd; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "<tr class=\"sd\">");
      fprintf(fp, "<th>%s</th><th></th><th></th>", "SD");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) switch(partype[j]) {
          case 0: fprintf(fp, "<th>%.0f</th>", p[j]); break;
          case 1: fprintf(fp, "<th>%.4f</th>", p[j]); break;
          default: fprintf(fp, "<th>%.4E</th>", p[j]); break;
        } else fprintf(fp, "<th></th>");
      }
      fprintf(fp, "</tr>\n");
    }
    /* Write lower confidence limits, if they exist */
    p=res->voi[i].cl1; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "<tr class=\"cl1\"");
      fprintf(fp, "<th>%s</th><th>%s</th><th>%s</th>", "CL", "95%", "Lower");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) switch(partype[j]) {
          case 0: fprintf(fp, "<th>%.0f</th>", p[j]); break;
          case 1: fprintf(fp, "<th>%.4f</th>", p[j]); break;
          default: fprintf(fp, "<th>%.4E</th>", p[j]); break;
        } else fprintf(fp, "<th></th>");
      }
      fprintf(fp, "</tr>\n");
    }
    /* Write upper confidence limits, if they exist */
    p=res->voi[i].cl2; for(j=n=0; j<res->parNr; j++) if(p[j]!=NA) n++;
    if(n>0) {
      fprintf(fp, "<tr class=\"cl2\"");
      fprintf(fp, "<th>%s</th><th>%s</th><th>%s</th>", "CL", "95%", "Upper");
      for(j=0; j<res->parNr; j++) {
        if(p[j]!=NA) switch(partype[j]) {
          case 0: fprintf(fp, "<th>%.0f</th>", p[j]); break;
          case 1: fprintf(fp, "<th>%.4f</th>", p[j]); break;
          default: fprintf(fp, "<th>%.4E</th>", p[j]); break;
        } else fprintf(fp, "<th></th>");
      }
      fprintf(fp, "</tr>\n");
    }
  }
  /* End the body of the table and table */
  fprintf(fp, "</tbody></table>\n");

  /* Stop writing the body of the HTML file, and end the file */
  fprintf(fp, "</div>\n");
  n=fprintf(fp, "</body></html>\n\n");
  if(n==0) {
    strcpy(reserrmsg, "disk full");
    if(!is_stdout) fclose(fp); return(3);
  }

  /* Close file */
  if(!is_stdout) fclose(fp);
  strcpy(reserrmsg, "");

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write XHTML 1.1 doctype into an opened file pointer.
\return Returns 0 if successful, non-zero in case of a failure.
 */
int resWriteXHTML11_doctype(
  FILE *fp
) {
  int n;
  if(fp==NULL) return(1);
  n=fprintf(fp, "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n");
  if(n<20) return(2);
  n=fprintf(fp, "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n\n");
  if(n<20) return(2);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Write XHTML 1.1 head for PET results file into an opened file pointer.
\return Returns 0 if successful, non-zero in case of a failure.
 */
int resWriteXHTML11_head(
  /** File pointer where to write */
  FILE *fp,
  /** Author name, for example software name */
  char *author_name
) {
  int n;
  if(fp==NULL) return(1);

  n=fprintf(fp, "<head>\n"); if(n<6) return(2);
  fprintf(fp, "  <title>Tabulated PET results</title>\n");
  fprintf(fp, "  <meta http-equiv=\"content-type\" content=\"text/html; charset=iso-8859-1\" />\n");
  fprintf(fp, "  <meta http-equiv=\"content-language\" content=\"en-gb\" />\n");
  fprintf(fp, "  <meta name=\"description\" content=\"Regional PET results\" />\n");
  fprintf(fp, "  <meta name=\"author\" content=\"%s\" />\n", author_name);
  fprintf(fp, "  <meta name=\"ProgId\" content=\"Excel.Sheet\" />\n");
  fprintf(fp, "  <link rel=\"icon\" href=\"http://www.turkupetcentre.net/favicon.ico\" type=\"image/x-icon\" />\n");
  fprintf(fp, "  <link rel=\"shortcut icon\" href=\"http://www.turkupetcentre.net/favicon.ico\" type=\"image/x-icon\" />\n");
  /* write local CSS with basic settings in case that external CSS is not available */
  fprintf(fp, "  <style type=\"text/css\">\n");
  fprintf(fp, "    thead {background-color:#999999; color:black;}\n");
  fprintf(fp, "    table {text-align:left; width:100%%; border-collapse:collapse; empty-cells:show;}\n");
  fprintf(fp, "    td {border:1px solid black;}\n");
  fprintf(fp, "    .oddroi {background-color:#FFFFFF; color:black;}\n");
  fprintf(fp, "    .evenroi {background-color:#CCCCCC; color:black;}\n");
  fprintf(fp, "    .sd {background-color:#999999; color:blue;}\n");
  fprintf(fp, "    .cl1 {background-color:#999999; color:green;}\n");
  fprintf(fp, "    .cl2 {background-color:#999999; color:green;}\n");
  fprintf(fp, "    .oddstudy {background-color:#FFFFFF; color:black;}\n");
  fprintf(fp, "    .evenstudy {background-color:#CCCCCC; color:black;}\n");
  fprintf(fp, "    .oddsum {background-color:#999999; color:black;}\n");
  fprintf(fp, "    .evensum {background-color:#CCCCCC; color:black;}\n");
  fprintf(fp, "    #regcontainer ul {margin-left:0; padding-left:0;}\n");
  fprintf(fp, "    #regcontainer ul li {display:inline; list-style-type:none;}\n");
  fprintf(fp, "    #regcontainer a {padding:2px 4px;}\n");
  fprintf(fp, "    <!--table\n");
  fprintf(fp, "    	{mso-displayed-decimal-separator:\"\\.\";\n");
  fprintf(fp, "    	mso-displayed-thousand-separator:\" \";}\n");
  fprintf(fp, "    -->\n");
  fprintf(fp, "  </style>\n");
  /* load external CSS with more fancy settings */
  fprintf(fp, "  <link rel=\"stylesheet\" type=\"text/css\" href=\"http://www.turkupetcentre.net/result.css\" />\n");
  fprintf(fp, "</head>\n");
  if(n<7) return(2);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** See studynr_from_fname() */
int resFName2study(char *fname, char *studyNumber)
{
  return(studynr_from_fname(fname, studyNumber));
}
/*****************************************************************************/

/*****************************************************************************/
/** Calculate the median and the lowest and highest value in the
    specified double array data of length nr.
    Note that array is sorted in this function.
    NULL pointer may be specified to function in place of an unwanted
    return parameter.
\return Returns 0 if succesfull.
 */
int resMedian(double *data, int nr, double *median, double *min, double *max)
{
  /* Check the arguments */
  if(data==NULL) return(1); if(nr<1) return(2);
  /* Sort data in increasing order */
  qsort(data, nr, sizeof(double), resQSortComp);
  /* Get minimum and maximum */
  if(min!=NULL) *min=data[0];
  if(max!=NULL) *max=data[nr-1];
  /* Calculate median */
  if(median!=NULL) {
    if(nr%2) *median=data[(nr-1)/2];
    else *median=0.5*(data[(nr/2)-1]+data[nr/2]);
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Calculate the mean and sd in the specified double array data of length nr.
    NULL pointer may be specified to function in place of an unwanted
    return parameter.
\return Returns 0 if succesfull.
 */
int resMean(double *data, int nr, double *mean, double *sd)
{
  int i;
  double sum, ssum, v;

  /* Check the arguments */
  if(data==NULL) return(1); if(nr<1) return(2);
  /* Calculate avg and sd */
  for(i=0, sum=ssum=0.0; i<nr; i++) {
    sum+=data[i]; ssum+=data[i]*data[i];
  }
  if(mean!=NULL) *mean=sum/(double)nr;
  if(sd!=NULL) {
    if(nr>1) v=(ssum-sum*sum/(double)nr)/(double)(nr-1); else v=0.0;
    if(v>0.0) *sd=sqrt(v); else *sd=0.0;
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
int resQSortComp(const void *f1, const void *f2)
{
  if(*(double*)f1<*(double*)f2) return(-1);
  else if(*(double*)f1>*(double*)f2) return(1);
  else return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Sort RES regions by region name. */
void resSortByName(RES *res)
{
  if(res==NULL || res->voiNr<=1) return;
  qsort(res->voi, res->voiNr, sizeof(ResVOI), resQSortName);
  return;
}
/*****************************************************************************/

/*****************************************************************************/
int resQSortName(const void *voi1, const void *voi2)
{
  int res;

  res=strcasecmp( ((const ResVOI*)voi1)->voiname, ((const ResVOI*)voi2)->voiname );
  if(res!=0) return(res);
  res=strcasecmp( ((const ResVOI*)voi1)->hemisphere, ((const ResVOI*)voi2)->hemisphere );
  if(res!=0) return(res);
  res=strcasecmp( ((const ResVOI*)voi1)->place, ((const ResVOI*)voi2)->place );
  return(res);
}
/*****************************************************************************/

/*****************************************************************************/
/** Copy result main header information to another result structure. */
int resCopyMHeader(RES *res1, RES *res2)
{
  if(res1==NULL || res2==NULL) return(1);
  strcpy(res2->program, res1->program);
  res2->time=res1->time;
  res2->parNr=res1->parNr;
  strcpy(res2->studynr, res1->studynr);
  strcpy(res2->datafile, res1->datafile);
  strcpy(res2->reffile, res1->reffile);
  strcpy(res2->plasmafile, res1->plasmafile);
  strcpy(res2->bloodfile, res1->bloodfile);
  strcpy(res2->refroi, res1->refroi);
  strcpy(res2->datarange, res1->datarange);
  res2->datanr=res1->datanr;
  strcpy(res2->fitmethod, res1->fitmethod);
  res2->density=res1->density;
  res2->lc=res1->lc;
  res2->concentration=res1->concentration;
  res2->beta=res1->beta;
  res2->Vb=res1->Vb;
  res2->isweight=res1->isweight;
  strcpy(res2->titleline, res1->titleline);
  strcpy(res2->unitline, res1->unitline);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Delete specified region (0..voiNr-1) from the structure.
\return Returns 0 if ok.
 */
int resDelete(RES *res, int voi)
{
  int i;

  /* Check that region exists */
  if(voi>res->voiNr-1 || voi<0) return(1);
  /* If it is the last one, then just decrease the voiNr */
  if(voi==res->voiNr) {res->voiNr--; return(0);}
  /* Then we have to move the following regions in its place */
  for(i=voi+1; i<res->voiNr; i++)
    memcpy(&res->voi[i-1], &res->voi[i], sizeof(ResVOI));
  res->voiNr--;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Select VOIs (sets sw=1), whose names are matching specified string.
    If no string is specified, then all VOIs are selected.
\return Returns the number of matches, or <0, if an error occurred.
 */
int resSelect(RES *data, char *name)
{
  int i, j, n;
  char *p, n1[128], n2[128], n3[128], tmp[128], sname[1024];

  /* Select all, if no string was specified */
  if(name==NULL || strlen(name)==0) {
    for(i=0; i<data->voiNr; i++) data->voi[i].sw=1; return(data->voiNr);}
  /* Make a copy of 'name' and use it */
  strcpy(sname, name);
  /* Check if string contains several substrings (hemisphere and place) */
  n1[0]=n2[0]=n3[0]=(char)0;
  p=strtok(sname, " ,;\n\t|"); if(p!=NULL) strcpy(n1, p); else return(-1);
  p=strtok(NULL, " ,;\n\t|"); if(p!=NULL) {
    strcpy(n2, p); p=strtok(NULL, " ,;\n\t|"); if(p!=NULL) strcpy(n3, p);}
  /* Convert strings to lowercase */
  for(i=0; i<strlen(n1); i++) n1[i]=tolower(n1[i]);
  for(i=0; i<strlen(n2); i++) n2[i]=tolower(n2[i]);
  for(i=0; i<strlen(n3); i++) n3[i]=tolower(n3[i]);
  /* Search through the data */
  for(i=0, n=0; i<data->voiNr; i++) {
    data->voi[i].sw=0;
    snprintf(tmp, 128, "%s%s%s", data->voi[i].voiname, data->voi[i].hemisphere,
                           data->voi[i].place);
    for(j=0; j<strlen(tmp); j++) tmp[j]=tolower(tmp[j]);
    if(strstr(tmp, n1)==NULL) continue;
    if(n2[0] && strstr(tmp, n2)==NULL) continue;
    if(n3[0] && strstr(tmp, n3)==NULL) continue;
    data->voi[i].sw=1; n++;
  }
  return(n);
}
/*****************************************************************************/

/*****************************************************************************/
/** Select the VOIs that have matching region name or number.
    Sets sw=1 or sw=0. This will replace resSelect().
\return Returns the number of selected VOIs, or <0 in case of an error.
 */
int resSelectRegions(
  /** Pointer to RES data where VOIs are selected */
  RES *res,
  /** Name or VOI number which is searched */
  char *region_name,
  /** 1=Non-matching VOIs are deselected, 0=Old selections are preserved */
  int reset
) {
  int ri, match_nr=0;

  /* Check the input */
  if(res==NULL || res->voiNr<1 || strlen(region_name)<1) return(-1);
  /* Reset all selections if required */
  if(reset!=0) for(ri=0; ri<res->voiNr; ri++) res->voi[ri].sw=0;
  /* Check each VOI */
  for(ri=0; ri<res->voiNr; ri++) {
    if(rnameMatch(res->voi[ri].name, ri+1, region_name)!=0) {
      res->voi[ri].sw=1; match_nr++;
    }
  }
  return(match_nr);
}
/*****************************************************************************/

/*****************************************************************************/
/** Determine whether the result parameter should be printed as
    integer (0), float (1), or exponential (2).
\return Returns the code, or <0 in case of an error.
 */
int resParameterPrintType(
  /** Pointer to the result struct */
  RES *res,
  /** Index of the parameter to test */
  int parIndex
) {
  int vi, partype=0;
  double x, m=0.0, pint;

  if(res==NULL || res->voiNr<1 || parIndex<0 || parIndex>=res->parNr)
    return(-1);
  for(vi=0; vi<res->voiNr; vi++) {
    x=res->voi[vi].parameter[parIndex];
    if(modf(x, &pint)!=0.0) partype=1;
    x=fabs(x); if(x>m) m=x;
  }
  if(partype==1 && (m>=10.0 || m<0.1)) partype=2;
  return(partype);
}
/*****************************************************************************/

/*****************************************************************************/
/** Check if result structure contains duplicate region names.
\return Returns 0 if not, 1 if duplicates are found, and <0 in case of an error.
 */  
int resIsDuplicateNames(RES *res)
{
  int ri, rj;
  
  if(res==NULL) return(-1);
  if(res->voiNr<2) return(0);
  for(ri=0; ri<res->voiNr-1; ri++) for(rj=ri+1; rj<res->voiNr; rj++)
    if(strcasecmp(res->voi[ri].name, res->voi[rj].name)==0) return(1);
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
